# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

import bpy
import os
import re
from mathutils import Vector, Matrix, Euler, Color, geometry
from math import pi, radians, sqrt

import bmesh
from .. import dxfgrabber
from . import convert, is_, groupsort
from .line_merger import line_merger
from ..transverse_mercator import TransverseMercator


try:
    from pyproj import Proj, transform as proj_transform
    PYPROJ = True
except:
    PYPROJ = False

BY_LAYER = 0
BY_DXFTYPE = 1
BY_CLOSED_NO_BULGE_POLY = 2
SEPARATED = 3
LINKED_OBJECTS = 4
GROUP_INSTANCES = 5
BY_BLOCK = 6


def transform(p1, p2, c1, c2, c3):
    if PYPROJ:
        if type(p1) is Proj and type(p2) is Proj:
            if p1.srs != p2.srs:
                return proj_transform(p1, p2, c1, c2, c3)
            else:
                return (c1, c2, c3)
        elif type(p2) is TransverseMercator:
            wgs84 = Proj(init="EPSG:4326")
            if p1.srs != wgs84.srs:
                t2, t1, t3 = proj_transform(p1, wgs84, c1, c2, c3)
            else:
                t1, t2, t3 = c2, c1, c3  # mind c2, c1 inversion
            tm1, tm2 = p2.fromGeographic(t1, t2)
            return (tm1, tm2, t3)
    else:
        if p1.spherical:
            t1, t2 = p2.fromGeographic(c2, c1)  # mind c2, c1 inversion
            return (t1, t2, c3)
        else:
            return (c1, c2, c3)


def float_len(f):
    s = str(f)
    if 'e' in s:
        return int(s[3:5])
    else:
        return len(s[2:])


class Indicator:
    spherical = False
    euclidean = False

    def __init__(self, i):
        if i.upper() == "SPHERICAL":
            self.spherical = True
        elif i.upper() == "EUCLIDEAN":
            self.euclidean = True
        else:
            raise AttributeError("Indication must be 'spherical' or 'euclidean'. Given: " + str(i))


class Do:
    __slots__ = (
        "dwg", "combination", "known_blocks", "import_text", "import_light", "export_acis", "merge_lines",
        "do_bounding_boxes", "acis_files", "errors", "block_representation", "recenter", "did_group_instance",
        "objects_before", "pDXF", "pScene", "thickness_and_width", "but_group_by_att", "current_scene",
        "dxf_unit_scale"
    )

    def __init__(self, dxf_filename, c=BY_LAYER, import_text=True, import_light=True, export_acis=True,
                 merge_lines=True, do_bbox=True, block_rep=LINKED_OBJECTS, recenter=False, pDXF=None, pScene=None,
                 thicknessWidth=True, but_group_by_att=True, dxf_unit_scale=1.0):
        self.dwg = dxfgrabber.readfile(dxf_filename, {"assure_3d_coords": True})
        self.combination = c
        self.known_blocks = {}
        self.import_text = import_text
        self.import_light = import_light
        self.export_acis = export_acis
        self.merge_lines = merge_lines
        self.do_bounding_boxes = do_bbox
        self.acis_files = []
        self.errors = set([])
        self.block_representation = block_rep
        self.recenter = recenter
        self.did_group_instance = False
        self.objects_before = []
        self.pDXF = pDXF
        self.pScene = pScene
        self.thickness_and_width = thicknessWidth
        self.but_group_by_att = but_group_by_att
        self.current_scene = None
        self.dxf_unit_scale = dxf_unit_scale

    def proj(self, co, elevation=0):
        """
        :param co: coordinate
        :param elevation: float (lwpolyline code 38)
        :return: transformed coordinate if self.pScene is defined
        """
        if self.pScene is not None and self.pDXF is not None:
            u = self.dxf_unit_scale
            if len(co) == 3:
                c1, c2, c3 = co
                c3 += elevation
            else:
                c1, c2 = co
                c3 = elevation
            if u != 1.0:
                c1 *= u
                c2 *= u
                c3 *= u
            # add
            add = Vector((0, 0, 0))
            if "latitude" in self.current_scene and "longitude" in self.current_scene:
                if PYPROJ and type(self.pScene) not in (TransverseMercator, Indicator):
                    wgs84 = Proj(init="EPSG:4326")
                    cscn_lat = self.current_scene.get('latitude', 0)
                    cscn_lon = self.current_scene.get('longitude', 0)
                    cscn_alt = self.current_scene.get('altitude', 0)
                    add = Vector(transform(wgs84, self.pScene, cscn_lon, cscn_lat, cscn_alt))

            # projection
            newco = Vector(transform(self.pDXF, self.pScene, c1, c2, c3))
            newco = newco - add
            if any((c == float("inf") or c == float("-inf") for c in newco)):
                self.errors.add("Projection results in +/- infinity coordinates.")
            return newco
        else:
            u = self.dxf_unit_scale
            if u != 1:
                if len(co) == 3:
                    c1, c2, c3 = co
                    c3 += elevation
                else:
                    c1, c2 = co
                    c3 = elevation
                c1 *= u
                c2 *= u
                c3 *= u
                return Vector((c1, c2, c3))
            else:
                return Vector((co[0], co[1], co[2] + elevation if len(co) == 3 else elevation))

    def georeference(self, scene, center):
        if "latitude" not in scene and "longitude" not in scene:
            if type(self.pScene) is TransverseMercator:
                scene['latitude'] = self.pScene.lat
                scene['longitude'] = self.pScene.lon
                scene['altitude'] = 0
            elif type(self.pScene) is not None:
                wgs84 = Proj(init="EPSG:4326")
                latlon = transform(self.pScene, wgs84, center[0], center[1], center[2])
                scene['longitude'] = latlon[0]
                scene['latitude'] = latlon[1]
                scene['altitude'] = latlon[2]

    """ GEOMETRY DXF TYPES TO BLENDER CURVES FILTERS"""
    # type(self, dxf entity, blender curve data)

    def _cubic_bezier_closed(self, ptuple, curve):
        count = (len(ptuple)-1)/3
        points = [ptuple[-2]]
        ptuples = ptuple[:-2]
        points += [p for p in ptuples]

        spl = curve.splines.new('BEZIER')
        spl.use_cyclic_u = True
        b = spl.bezier_points
        b.add(count - 1)
        for i, j in enumerate(range(1, len(points), 3)):
            b[i].handle_left = self.proj(points[j - 1])
            b[i].co = self.proj(points[j])
            b[i].handle_right = self.proj(points[j + 1])

    def _cubic_bezier_open(self, points, curve):
        count = (len(points) - 1) / 3 + 1
        spl = curve.splines.new('BEZIER')
        b = spl.bezier_points
        b.add(count - 1)

        b[0].co = self.proj(points[0])
        b[0].handle_left = self.proj(points[0])
        b[0].handle_right = self.proj(points[1])

        b[-1].co = self.proj(points[-1])
        b[-1].handle_right = self.proj(points[-1])
        b[-1].handle_left = self.proj(points[-2])

        for i, j in enumerate(range(3, len(points) - 2, 3), 1):
            b[i].handle_left = self.proj(points[j - 1])
            b[i].co = self.proj(points[j])
            b[i].handle_right = self.proj(points[j + 1])

    def _cubic_bezier(self, points, curve, is_closed):
        """
        points: control points; list of (x,y,z) tuples
        curve: Blender curve data of type "CURVE" (object.data) where the bezier should be added to
        is_closed: True / False to indicate if the curve is open or closed
        """
        if is_closed:
            self._cubic_bezier_closed(points, curve)
        else:
            self._cubic_bezier_open(points, curve)

    def _poly(self, points, curve, elevation=0, is_closed=False):
        """
        points: list of (x,y,z)
        curve: Blender curve data of type "CURVE" (object.data) to which the poly should be added to
        param elevation: float (lwpolyline code 38)
        is_closed: True / False to indicate if the polygon is open or closed
        """
        p = curve.splines.new("POLY")
        p.use_smooth = False
        p.use_cyclic_u = is_closed
        p.points.add(len(points) - 1)

        for i, pt in enumerate(points):
            p.points[i].co = self.proj(pt, elevation).to_4d()

    def _gen_poly(self, en, curve, elevation=0):
        if any([b != 0 for b in en.bulge]):
            self._cubic_bezier(convert.bulgepoly_to_cubic(self, en), curve, en.is_closed)
        else:
            self._poly(en.points, curve, elevation, en.is_closed)

    def polyline(self, en, curve):
        """
        en: DXF entity of type `POLYLINE`
        curve: Blender data structure of type `CURVE`
        """
        self._gen_poly(en, curve)

    def polygon(self, en, curve):
        """
        en: DXF entity of type `POLYGON`
        curve: Blender data structure of type `CURVE`
        """
        self._gen_poly(en, curve)

    def lwpolyline(self, en, curve):
        """
        en: DXF entity of type `LWPOLYLINE`
        curve: Blender data structure of type `CURVE`
        """
        self._gen_poly(en, curve, en.elevation)

    def line(self, en, curve):
        """
        en: DXF entity of type `LINE`
        curve: Blender data structure of type `CURVE`
        """
        self._poly([en.start, en.end], curve, 0, False)

    def arc(self, en, curve=None, aunits=None, angdir=None, angbase=None):
        """
        en: dxf entity (en.start_angle, en.end_angle, en.center, en.radius)
        curve: optional; Blender curve data of type "CURVE" (object.data) to which the arc should be added to
        return control points of a cubic spline (do be used in a spline with bulges / series of arcs)
        note: en.start_angle + en.end_angle: angles measured from the angle base (angbase) in the direction of
              angdir (1 = clockwise, 0 = counterclockwise)
        """
        threshold = 0.005

        if aunits is None:
            aunits = self.dwg.header.get('$AUNITS', 0)
        if angbase is None:
            angbase = self.dwg.header.get('$ANGBASE', 0)
        if angdir is None:
            angdir = self.dwg.header.get('$ANGDIR', 0)
        kappa = 0.5522848

        # TODO: add support for 1 (dms) and 4 (survey)
        if aunits == 0:
            # Degree
            s = radians(en.start_angle+angbase)
            e = radians(en.end_angle+angbase)
        elif aunits == 2:
            # Gradians
            s = radians(0.9 * (en.start_angle + angbase))
            e = radians(0.9 * (en.end_angle + angbase))
        else:
            # Radians
            s = en.start_angle+angbase
            e = en.end_angle+angbase

        if s > e:
            e += 2 * pi
        angle = e - s

        vc = Vector(en.center)
        x_vec = Vector((1, 0, 0))
        radius = en.radius

        # turn clockwise
        if angdir == 0:
            rot = Matrix.Rotation(radians(-90), 3, "Z")
            start = x_vec * radius @ Matrix.Rotation(-s, 3, "Z")
            end = x_vec * radius @ Matrix.Rotation(-e, 3, "Z")

        # turn counter-clockwise
        else:
            rot = Matrix.Rotation(radians(90), 3, "Z")
            start = x_vec * radius @ Matrix.Rotation(s, 3, "Z")
            end = x_vec * radius @ Matrix.Rotation(e, 3, "Z")

        # start
        spline = list()
        spline.append(vc + start)
        if abs(angle) - pi / 2 > threshold:  # if angle is more than pi/2 incl. threshold
            spline.append(vc + start + start * kappa @ rot)
        else:
            spline.append(vc + start + start * kappa * angle / (pi / 2) @ rot)

        # fill if angle is larger than 90 degrees
        a = pi / 2
        if abs(angle) - threshold > a:
            fill = start

            while abs(angle) - a > threshold:
                fillnext = fill @ rot
                spline.append(vc + fillnext + fill * kappa)
                spline.append(vc + fillnext)
                # if this was the last fill control point
                if abs(angle) - a - pi / 2 < threshold:
                    end_angle = (abs(angle) - a) * abs(angle) / angle
                    spline.append(vc + fillnext + fillnext * kappa * end_angle / (pi / 2) @ rot)
                else:
                    spline.append(vc + fillnext + fillnext * kappa @ rot)
                fill = fillnext
                a += pi / 2

        else:
            end_angle = angle

        # end
        spline.append(vc + end + end * -kappa * end_angle / (pi / 2) @ rot)
        spline.append(vc + end)

        if len(spline) % 3 != 1:
            print("DXF-IMPORT: DO ARC: CHECK PLEASE: ", len(spline), spline)

        # curve is None means arc is called from bulge conversion
        # nothing should be projected at this stage, since the
        # lwpolyline (the only entity with bulges) will be projected
        # as a whole afterwards (small little error; took ages to debug)
        if curve is not None:
            self._cubic_bezier_open(spline, curve)
            return spline
        else:
            return spline

    def circle(self, en, curve, major=Vector((1, 0, 0))):
        """
        en: dxf entity
        curve: Blender curve data of type "CURVE" (object.data) to which the circle should be added to
        major: optional; if the circle is used as a base for an ellipse, major denotes the ellipse's major direction
        """
        c = curve.splines.new("BEZIER")
        c.use_cyclic_u = True
        b = c.bezier_points
        b.add(3)

        for i in range(4):
            b[i].handle_left_type = 'AUTO'
            b[i].handle_right_type = 'AUTO'

        vc = Vector(en.center)
        clockwise = Matrix(((0, -1, 0), (1, 0, 0), (0, 0, 1)))

        r = major
        if len(r) == 2:
            r = r.to_3d()
        r = r * en.radius

        try:
            b[0].co = self.proj(vc + r)
            b[1].co = self.proj(vc + r @ clockwise)
            b[2].co = self.proj(vc + r @ clockwise @ clockwise)
            b[3].co = self.proj(vc + r @ clockwise @ clockwise @ clockwise)
        except:
            print("Circle center: ", vc, "radius: ", r)
            raise

        return c

    def scale_controlpoint(self, p, factor):
        """
        p: Blender control-point
        factor: (Float)
        Repositions left and right handle of a bezier control point.
        """
        p.handle_left = p.co + (p.handle_left - p.co) * factor
        p.handle_right = p.co + (p.handle_right - p.co) * factor

    def ellipse(self, en, curve):
        """
        center: (x,y,z) of circle center
        major: the ellipse's major direction
        ratio: ratio between major and minor axis lengths (always < 1)
        curve: Blender curve data of type "CURVE" (object.data) to which the ellipse should be added to
        """
        major = Vector(en.major_axis)
        en.__dict__["radius"] = major.length
        c = self.circle(en, curve, major.normalized())
        b = c.bezier_points

        if en.ratio < 1:
            for i in range(4):
                b[i].handle_left_type = 'ALIGNED'
                b[i].handle_right_type = 'ALIGNED'

            vc = self.proj(en.center)
            clockwise = Matrix(((0, -1, 0), (1, 0, 0), (0, 0, 1)))
            if len(major) == 2:
                major = major.to_3d()
            minor = major * en.ratio @ clockwise

            lh = b[1].handle_left - b[1].co
            rh = b[1].handle_right - b[1].co
            b[1].co = vc + minor
            b[1].handle_left = b[1].co + lh
            b[1].handle_right = b[1].co + rh
            b[3].co = vc + minor @ clockwise @ clockwise
            b[3].handle_left = b[3].co + rh
            b[3].handle_right = b[3].co + lh

            self.scale_controlpoint(b[0], en.ratio)
            self.scale_controlpoint(b[2], en.ratio)

    def spline(self, en, curve, _3D=True):
        """
        en: DXF entity of type `SPLINE`
        curve: Blender data structure of type `CURVE`
        """
        if _3D:
            curve.dimensions = "3D"
        spline = convert.bspline_to_cubic(self, en, curve, self.errors)
        if spline is None:
            self.errors.add("Not able to import bspline with degree > 3")
        else:
            self._cubic_bezier(spline, curve, en.is_closed)

    def helix(self, en, curve):
        """
        en: DXF entity of type `HELIX`
        curve: Blender data structure of type `CURVE`
        """
        self.spline(en, curve, not en.is_planar)

    """ GEOMETRY DXF TYPES TO BLENDER MESHES FILTERS"""
    # type(self, dxf entity, blender bmesh data)

    def _gen_meshface(self, points, bm):
        """
        points: list of (x,y,z) tuples
        bm: bmesh to add the (face-) points
        Used by the3dface() and solid()
        """

        def _is_on_edge(point):
            return abs(sum((e - point).length for e in (edge1, edge2)) - (edge1 - edge2).length) < 0.01

        points = list(points)

        i = 0
        while i < len(points):
            if points.count(points[i]) > 1:
                points.pop(i)
            else:
                i += 1

        verts = []
        for p in points:
            verts.append(bm.verts.new(self.proj(p)))

        # add only an edge if len points < 3
        if len(points) == 2:
            bm.edges.new(verts)
        elif len(points) > 2:
            face = bm.faces.new(verts)

            if len(points) == 4:
                for i in range(2):
                    edge1 = verts[i].co
                    edge2 = verts[i + 1].co
                    opposite1 = verts[i + 2].co
                    opposite2 = verts[(i + 3) % 4].co
                    ii = geometry.intersect_line_line(edge1, edge2, opposite1, opposite2)
                    if ii is not None:
                        if _is_on_edge(ii[0]):
                            try:
                                bm.faces.remove(face)
                            except Exception as e:
                                pass
                            iv = bm.verts.new(ii[0])
                            bm.faces.new((verts[i], iv, verts[(i + 3) % 4]))
                            bm.faces.new((verts[i + 1], iv, verts[i + 2]))

    def the3dface(self, en, bm):
        """ f: dxf entity
            bm: Blender bmesh data to which the 3DFACE should be added to.
            <-? bm.from_mesh(object.data) ?->
        """
        if en.points[-1] == en.points[-2]:
            points = en.points[:3]
        else:
            points = en.points
        self._gen_meshface(points, bm)

    def solid(self, en, bm):
        """ f: dxf entity
            bm: Blender bmesh data to which the SOLID should be added to.
            <-? bm.from_mesh(object.data) ?->
        """
        p = en.points
        points = (p[0], p[1], p[3], p[2])
        self._gen_meshface(points, bm)

    def trace(self, en, bm):
        self.solid(en, bm)

    def point(self, en, bm):
        """
        en: DXF entity of type `POINT`
        bm: Blender bmesh instance
        """
        bm.verts.new(en.point)

    def polyface(self, en, bm):
        """
        pf: polyface
        bm: Blender bmesh data to which the POLYFACE should be added to.
        <-? bm.from_mesh(object.data) ?->
        """
        for v in en.vertices:
            bm.verts.new(v.location)

        bm.verts.ensure_lookup_table()
        for subface in en:
            idx = subface.indices()
            points = []
            for p in idx:
                if p not in points:
                    points.append(p)
            if len(points) in (3, 4):
                bm.faces.new([bm.verts[i] for i in points])

    def polymesh(self, en, bm):
        """
        en: POLYMESH entity
        bm: Blender bmesh instance
        """
        mc = en.mcount if not en.is_mclosed else en.mcount + 1
        nc = en.ncount if not en.is_nclosed else en.ncount + 1
        for i in range(1, mc):
            i = i % en.mcount
            i_ = (i - 1) % en.mcount
            for j in range(1, nc):
                j = j % en.ncount
                j_ = (j - 1) % en.ncount
                face = []
                face.append(bm.verts.new(en.get_location((i_, j_))))
                face.append(bm.verts.new(en.get_location((i, j_))))
                face.append(bm.verts.new(en.get_location((i, j))))
                face.append(bm.verts.new(en.get_location((i_, j))))

                bm.faces.new(face)

    def mesh(self, en, bm):
        """
        mesh: dxf entity
        m: Blender MESH data (object.data) to which the dxf-mesh should be added
        """
        # verts:
        for v in en.vertices:
            bm.verts.new(v)

        # edges:
        bm.verts.ensure_lookup_table()
        if any((c < 0 for c in en.edge_crease_list)):
            layerkey = bm.edges.layers.crease.new("SubsurfCrease")
            for i, edge in enumerate(en.edges):
                bme = bm.edges.new([bm.verts[edge[0]], bm.verts[edge[1]]])
                bme[layerkey] = -en.edge_crease_list[i]
        else:
            for i, edge in enumerate(en.edges):
                bm.edges.new([bm.verts[edge[0]], bm.verts[edge[1]]])

        # faces:
        for face in en.faces:
            bm.faces.new([bm.verts[i] for i in face])

    """ SEPARATE BLENDER OBJECTS FROM (CON)TEXT / STRUCTURE DXF TYPES """
    # type(self, dxf entity, name string)
    #     returns blender object

    def _extrusion(self, obj, entity):
        """
        extrusion describes the normal vector of the entity
        """
        if entity.dxftype not in {"LINE", "POINT"}:
            if is_.extrusion(entity):
                transformation = convert.extrusion_to_matrix(entity)
                obj.location = transformation @ obj.location
                obj.rotation_euler.rotate(transformation)

    def _bbox(self, objects):
        xmin = ymin = zmin = float('+inf')
        xmax = ymax = zmax = float('-inf')
        bpy.context.view_layer.update()

        for obj in objects:
            om = obj.matrix_basis
            for v in obj.bound_box:
                p = om @ Vector(v)
                if p.x < xmin:
                    xmin = p.x
                if p.x > xmax:
                    xmax = p.x
                if p.y < ymin:
                    ymin = p.y
                if p.y > ymax:
                    ymax = p.y
                if p.z < zmin:
                    zmin = p.z
                if p.z > zmax:
                    zmax = p.z

        if xmin == float('+inf'):
            xmin = 0
        if ymin == float('+inf'):
            ymin = 0
        if zmin == float('+inf'):
            zmin = 0
        if xmax == float('-inf'):
            xmax = 0
        if ymax == float('-inf'):
            ymax = 0
        if zmax == float('-inf'):
            zmax = 0

        return xmin, ymin, zmin, xmax, ymax, zmax

    def _object_bbox(self, objects, name, do_widgets=True):
        xmin, ymin, zmin, xmax, ymax, zmax = self._bbox(objects)

        # creating bbox geometry
        bm = bmesh.new()
        verts = []
        # left side vertices
        verts.append(bm.verts.new((xmin, ymin, zmin)))
        verts.append(bm.verts.new((xmin, ymin, zmax)))
        verts.append(bm.verts.new((xmin, ymax, zmax)))
        verts.append(bm.verts.new((xmin, ymax, zmin)))
        # right side vertices
        verts.append(bm.verts.new((xmax, ymin, zmin)))
        verts.append(bm.verts.new((xmax, ymin, zmax)))
        verts.append(bm.verts.new((xmax, ymax, zmax)))
        verts.append(bm.verts.new((xmax, ymax, zmin)))

        # creating the widgets
        if do_widgets:
            for i in range(2):
                for j in range(4):
                    point = verts[j + (i * 4)]
                    prevp = verts[((j - 1) % 4) + (i * 4)]
                    nextp = verts[((j + 1) % 4) + (i * 4)]
                    neigh = verts[(j + (i * 4) + 4) % 8]

                    for con in (prevp, nextp, neigh):
                        vec = Vector(con.co - point.co) / 10
                        bm.edges.new((point, bm.verts.new(point.co + vec)))

        d = bpy.data.meshes.new(name + "BBOX")
        bm.to_mesh(d)
        o = bpy.data.objects.new(name, d)
        return o

    def _vertex_duplication(self, x, y, x_count, y_count):
        bm = bmesh.new()
        for i in range(x_count):
            for j in range(y_count):
                bm.verts.new(Vector((x * i, y * j, 0.)))

        d = bpy.data.meshes.new("vertex_duplicator")
        bm.to_mesh(d)
        return d

    def point_object(self, en, scene, name=None):
        if name is None:
            name = en.dxftype
        o = bpy.data.objects.new("Point", None)
        o.location = self.proj(en.point)
        self._extrusion(o, en)
        scene.collection.objects.link(o)

        group = self._get_group(en.layer)
        group.objects.link(o)

    def light(self, en, scene, name):
        """
        en: dxf entity
        name: ignored; exists to make separate and merged objects methods universally callable from _call_types()
        Creates, links and returns a new light object depending on the type and color of the dxf entity.
        """
        # light_type : distant = 1; point = 2; spot = 3
        if self.import_light:
            type_map = ["NONE", "SUN", "POINT", "SPOT"]
            layer = self.dwg.layers[en.layer]
            lamp = bpy.data.lights.new(en.name, type_map[en.light_type])
            if en.color != 256:
                aci = en.color
            else:
                aci = layer.color
            c = dxfgrabber.aci_to_true_color(aci)
            lamp.color = Color(c.rgb())
            if en.light_type == 3:
                lamp.spot_size = en.hotspot_angle
            o = bpy.data.objects.new(en.name, lamp)
            o.location = self.proj(en.position)
            dir = self.proj(en.target) - self.proj(en.position)
            o.rotation_quaternion = dir.rotation_difference(Vector((0, 0, -1)))
            scene.collection.objects.link(o)
            return o

    def mtext(self, en, scene, name):
        """
        en: dxf entity
        name: ignored; exists to make separate and merged objects methods universally callable from _call_types()
        Returns a new multi-line text object.
        """
        if self.import_text:
            text = en.plain_text()
            name = text[:8]
            d = bpy.data.curves.new(name, "FONT")
            o = bpy.data.objects.new(name, d)
            d.body = text
            d.size = en.height
            if en.rect_width is not None:
                if en.rect_width > 50:
                    width = 50
                    ratio = 50 / en.rect_width
                    d.size = en.height * ratio * 1.4  # XXX HACK
                    scale = (1 / ratio, 1 / ratio, 1 / ratio)
                    d.space_line = en.line_spacing
                else:
                    width = en.rect_width
                    scale = (1, 1, 1)
                d.text_boxes[0].width = width
                o.scale = scale

            # HACK
            d.space_line *= 1.5

            o.rotation_euler = Vector((1, 0, 0)).rotation_difference(en.xdirection).to_euler()
            o.location = en.insert
            return o

    def text(self, en, scene, name):
        """
        en: dxf entity
        name: ignored; exists to make separate and merged objects methods universally callable from _call_types()
        Returns a new single line text object.
        """
        if self.import_text:
            name = en.text[:8]
            d = bpy.data.curves.new(name, "FONT")
            d.body = en.plain_text()
            d.size = en.height
            o = bpy.data.objects.new(name, d)
            o.rotation_euler = Euler((0, 0, radians(en.rotation)), 'XYZ')
            basepoint = self.proj(en.basepoint) if hasattr(en, "basepoint") else self.proj((0, 0, 0))
            o.location = self.proj((en.insert)) + basepoint
            if hasattr(en, "thickness"):
                et = en.thickness / 2
                d.extrude = abs(et)
                if et > 0:
                    o.location.z += et
                elif et < 0:
                    o.location.z -= et
            return o

    def block_linked_object(self, entity, scene, name=None, override_group=None, invisible=None, recursion_level=0):
        """
        entity: DXF entity of type `BLOCK`
        name: to be consistent with all functions that get mapped to a DXF type; is set when called from insert()
        override_group: is set when called from insert() (bpy_types.Group)
        invisible: boolean to control the visibility of the returned object
        Returns an object. All blocks with the same name share the same geometry (Linked Blender Objects). If the
        entities in a block have to mapped to different Blender-types (like curve, mesh, surface, text, light) a list of
        sub-objects is being created. `INSERT`s inside a block are being added to the same list. If the list has more
        than one element they are being parented to a newly created Blender-Empty which is return instead of the
        single object in the sub-objects-list that is being returned only if it is the only object in the list.
        """
        def _recursive_copy_inserts(parent, known_inserts, inserts, group, invisible):
            for ki in known_inserts:
                new_insert = ki.copy()
                _recursive_copy_inserts(new_insert, ki.children, None, group, invisible)
                if new_insert.name not in group.objects:
                    group.objects.link(new_insert)
                if invisible is not None:
                    new_insert.hide_viewport = bool(invisible)
                if inserts is not None:
                    inserts.append(new_insert)
                new_insert.parent = parent
                scene.collection.objects.link(new_insert)

        if name is None:
            name = entity.name
        # get group
        if override_group is not None:
            group = override_group
        else:
            group = self._get_group(entity.layer)

        # get object(s)
        objects = []
        inserts = []
        if name not in self.known_blocks.keys():
            block_inserts = [en for en in entity if is_.insert(en.dxftype)]
            bc = (en for en in entity if is_.combined_entity(en))
            bs = (en for en in entity if is_.separated_entity(en) and not is_.insert(en.dxftype))

            if self.combination != SEPARATED:
                objects += self.combined_objects(bc, scene, "BL|" + name, group)
            else:
                bs = (en for en in entity if (is_.combined_entity(en) or is_.separated_entity(en)) and
                      not is_.insert(en.dxftype))
            objects += self.separated_entities(bs, scene, "BL|" + name, group)

            # create inserts - RECURSION
            insert_bounding_boxes = []
            for INSERT in block_inserts:
                insert = self.insert(INSERT, scene, None, group, invisible, recursion_level + 1)
                if len(insert.children) > 0:
                    i_copy = bpy.data.objects.new(insert.name, None)
                    i_copy.matrix_basis = insert.matrix_basis
                    scene.collection.objects.link(i_copy)
                    group.objects.link(i_copy)
                    kids = insert.children[:]
                    for child in kids:
                        child.parent = i_copy
                    inserts.append(i_copy)
                    insert_bounding_boxes.append(insert)
                else:
                    inserts.append(insert)

            # determining the main object o
            if len(objects) > 1 or len(insert_bounding_boxes) > 0:
                if self.do_bounding_boxes:
                    o = self._object_bbox(objects + insert_bounding_boxes, name, recursion_level == 0)
                    scene.collection.objects.link(o)
                else:
                    o = bpy.data.objects.new(name, None)
                    scene.collection.objects.link(o)
                if len(objects) > 0:
                    for obj in objects:
                        obj.parent = o

            elif len(objects) == 1:
                o = objects.pop(0)
            else:
                # strange case but possible according to the testfiles
                o = bpy.data.objects.new(name, None)
                scene.collection.objects.link(o)

            # unlink bounding boxes of inserts
            for ib in insert_bounding_boxes:
                if ib.name in group.objects:
                    group.objects.unlink(ib)
                scene.collection.objects.unlink(ib)

            # parent inserts to this block before any transformation on the block is being applied
            for obj in inserts:
                obj.parent = o

            # put a copy of the retrieved objects into the known_blocks dict, so that the attributes being added to
            # the object from this point onwards (from INSERT attributes) are not being copied to new/other INSERTs
            self.known_blocks[name] = [[o.copy() for o in objects], inserts]

            # so that it gets assigned to the group and inherits visibility too
            objects.append(o)

            self.known_blocks[name].append(o.copy())
        else:
            known_objects, known_inserts, known_o = self.known_blocks[name]

            for known_object in known_objects:
                oc = known_object.copy()
                scene.collection.objects.link(oc)
                objects.append(oc)

            o = known_o.copy()
            scene.collection.objects.link(o)

            _recursive_copy_inserts(o, known_inserts, inserts, group, invisible)

            # parent objects to o
            for obj in objects:
                obj.parent = o

            objects.append(o)

        # link group
        for obj in objects:
            if obj.name not in group.objects:
                group.objects.link(obj)

        # visibility
        if invisible is not None:
            for obj in objects:
                obj.hide_viewport = bool(invisible)

        # block transformations
        o.location = self.proj(entity.basepoint)

        return o

    def block_group_instances(self, entity, scene, name=None, override_group=None, invisible=None, recursion_level=0):
        self.did_group_instance = True

        if name is None:
            name = entity.name
        # get group
        if override_group is not None:
            group = override_group
        else:
            group = self._get_group(entity.layer)

        block_group = self._get_group(entity.name+"_BLOCK")

        if "Blocks" not in bpy.data.scenes:
            block_scene = bpy.data.scenes.new("Blocks")
        else:
            block_scene = bpy.data.scenes["Blocks"]

        # create the block
        if len(block_group.objects) == 0 or name not in self.known_blocks.keys():
            bpy.context.screen.scene = block_scene
            block_inserts = [en for en in entity if is_.insert(en.dxftype)]
            bc = (en for en in entity if is_.combined_entity(en))
            bs = (en for en in entity if is_.separated_entity(en) and not is_.insert(en.dxftype))

            objects = []
            if self.combination != SEPARATED:
                objects += self.combined_objects(bc, block_scene, "BL|" + name, block_group)
            else:
                bs = (en for en in entity if (is_.combined_entity(en) or is_.separated_entity(en)) and
                      not is_.insert(en.dxftype))
            objects += self.separated_entities(bs, block_scene, "BL|" + name, block_group)

            # create inserts - RECURSION
            inserts = []
            for INSERT in block_inserts:
                i = self.insert(INSERT, block_scene, None, block_group, invisible, recursion_level + 1, True)
                inserts.append(i)

            bbox = self._object_bbox(objects + inserts, name, True)

            for i in inserts:
                sub_group = i.instance_collection
                block_scene.collection.objects.unlink(i)
                block_group.objects.unlink(i)
                i_empty = bpy.data.objects.new(i.name, None)
                i_empty.matrix_basis = i.matrix_basis
                i_empty.instance_type = "COLLECTION"
                i_empty.instance_collection = sub_group
                block_group.objects.link(i_empty)
                block_scene.collection.objects.link(i_empty)

            self.known_blocks[name] = [objects, inserts, bbox]
        else:
            bbox = self.known_blocks[name][2]

        bpy.context.screen.scene = scene
        o = bbox.copy()
        # o.empty_display_size = 0.3
        o.instance_type = "COLLECTION"
        o.instance_collection = block_group
        group.objects.link(o)
        if invisible is not None:
            o.hide_viewport = invisible
        o.location = self.proj(entity.basepoint)
        scene.collection.objects.link(o)
        # block_scene.view_layers[0].update()

        return o

    def insert(self, entity, scene, name, group=None, invisible=None, recursion_level=0, need_group_inst=None):
        """
        entity: DXF entity
        name: String; not used but required to be consistent with the methods being called from _call_type()
        group: Blender group of type (bpy_types.Collection) being set if called from block()
        invisible: boolean to control visibility; being set if called from block()
        """
        aunits = self.dwg.header.get('$AUNITS', 0)

        # check if group instances are needed
        kids = sum(1 for i in self.dwg.blocks[entity.name] if i.dxftype == "INSERT")
        sep = sum(1 for sep in self.dwg.blocks[entity.name] if is_.separated_entity(sep))
        objtypes = sum(1 for ot, ens in groupsort.by_blender_type(en for en in self.dwg.blocks[entity.name]
                                                                  if is_.combined_entity(en))
                       if ot in {"object_mesh", "object_curve"})
        if need_group_inst is None:
            need_group_inst = (entity.row_count or entity.col_count) > 1 and \
                              (kids > 0 or objtypes > 1 or sep > 1 or (objtypes > 0 and sep > 0))

        if group is None:
            group = self._get_group(entity.layer)

        if self.block_representation == GROUP_INSTANCES or need_group_inst:
            o = self.block_group_instances(self.dwg.blocks[entity.name], scene, entity.name, group,
                                           entity.invisible, recursion_level)
        else:
            o = self.block_linked_object(self.dwg.blocks[entity.name], scene, entity.name, group,
                                         entity.invisible, recursion_level)

        # column & row
        if (entity.row_count or entity.col_count) > 1:
            if len(o.children) == 0 and self.block_representation == LINKED_OBJECTS and not need_group_inst:
                    arr_row = o.modifiers.new("ROW", "ARRAY")
                    arr_col = o.modifiers.new("COL", "ARRAY")
                    arr_row.show_expanded = False
                    arr_row.count = entity.row_count
                    arr_row.relative_offset_displace = (0, entity.row_spacing, 0)
                    arr_col.show_expanded = False
                    arr_col.count = entity.col_count
                    arr_col.relative_offset_displace = (entity.col_spacing / 2, 0, 0)

            else:
                instance = o
                x = (Vector(o.bound_box[4]) - Vector(o.bound_box[0])).length
                y = (Vector(o.bound_box[3]) - Vector(o.bound_box[0])).length
                dm = self._vertex_duplication(x * entity.col_spacing / 2, y * entity.row_spacing,
                                              entity.col_count, entity.row_count)
                o = bpy.data.objects.new(entity.name, dm)
                instance.parent = o
                o.instance_type = "VERTS"

        # insert transformations
        rot = radians(entity.rotation) if aunits == 0 else entity.rotation
        o.location += self.proj(entity.insert)
        o.rotation_euler = Euler((0, 0, rot))
        o.scale = entity.scale

        # mirror (extrusion value of an INSERT ENTITY)
        self._extrusion(o, entity)

        # visibility
        if invisible is None:
            o.hide_viewport = bool(entity.invisible)
        else:
            o.hide_viewport = bool(invisible)

        # attributes
        if self.import_text:
            if entity.attribsfollow:
                for a in entity.attribs:
                    # Blender custom property
                    o[a.tag] = a.text
                    attname = entity.name + "_" + a.tag
                    scene.collection.objects.link(self.text(a, scene, attname))

        return o

    """ COMBINED BLENDER OBJECT FROM GEOMETRY DXF TYPES """
    # type(self, dxf entities, object name string)
    #     returns blender object

    def _check3D_object(self, curve):
        """
        Checks if a curve object has coordinates that are elevated from the z-plane.
        If so the curve.dimensions are set to 3D.
        """
        if any((p.co.z != 0 for spline in curve.splines for p in spline.points)) or \
           any((p.co.z != 0 for spline in curve.splines for p in spline.bezier_points)):
            curve.dimensions = '3D'

    def _merge_lines(self, lines, curve):
        """
        lines: list of LINE entities
        curve: Blender curve data
        merges a list of LINE entities to a polygon-point-list and adds it to the Blender curve
        """
        polylines = line_merger(lines)
        for polyline in polylines:
            self._poly(polyline, curve, 0, polyline[0] == polyline[-1])

    def _thickness(self, bm, thickness):
        """
        Used for mesh types
        """
        if not self.thickness_and_width:
            return
        original_faces = [face for face in bm.faces]
        bm.normal_update()
        for face in original_faces:
            normal = face.normal.copy()
            if normal.z < 0:
                normal *= -1
            ret = bmesh.ops.extrude_face_region(bm, geom=[face])
            new_geom = ret["geom"]
            verts = (g for g in new_geom if type(g) == bmesh.types.BMVert)
            for v in verts:
                v.co += normal * thickness
            del ret
        del original_faces

    def _thickness_and_width(self, obj, entity, scene):
        """
        Used for curve types
        """
        if not self.thickness_and_width:
            return
        has_varying_width = is_.varying_width(entity)
        th = entity.thickness
        w = 0
        if hasattr(entity, "width"):
            if len(entity.width) > 0 and len(entity.width[0]) > 0:
                w = entity.width[0][0]

        if w == 0 and not has_varying_width:
            if th != 0:
                obj.data.extrude = abs(th / 2)
                if th > 0:
                    obj.location.z += th / 2
                else:
                    obj.location.z -= th / 2
            obj.data.dimensions = "3D"
            obj.data.twist_mode = "Z_UP"

        else:
            # CURVE BEVEL
            ew = entity.width
            max_w = max((w for w_pair in ew for w in w_pair))

            bevd = bpy.data.curves.new("BEVEL", "CURVE")
            bevdp = bevd.splines.new("POLY")
            bevdp.points.add(1)
            bevdp.points[0].co = Vector((-max_w / 2, 0, 0, 0))
            bevdp.points[1].co = Vector((max_w / 2, 0, 0, 0))

            bevel = bpy.data.objects.new("BEVEL", bevd)
            obj.data.bevel_object = bevel
            scene.collection.objects.link(bevel)

            # CURVE TAPER
            if has_varying_width and len(ew) == 1:
                tapd = bpy.data.curves.new("TAPER", "CURVE")
                tapdp = tapd.splines.new("POLY")
                # lenlist = convert.bulgepoly_to_lenlist(entity)
                # amount = len(ew) if entity.is_closed else len(ew) - 1

                tapdp.points[0].co = Vector((0, ew[0][0] / max_w, 0, 0))
                tapdp.points.add(1)
                tapdp.points[1].co = Vector((1, ew[0][1] / max_w, 0, 0))

                # for i in range(1, amount):
                #     start_w = ew[i][0]
                #     end_w = ew[i][1]
                #     tapdp.points.add(2)
                #     tapdp.points[-2].co = Vector((sum(lenlist[:i]), start_w / max_w, 0, 0))
                #     tapdp.points[-1].co = Vector((sum(lenlist[:i + 1]), end_w / max_w, 0, 0))

                taper = bpy.data.objects.new("TAPER", tapd)
                obj.data.taper_object = taper
                scene.collection.objects.link(taper)

            # THICKNESS FOR CURVES HAVING A WIDTH
            if th != 0:
                solidify = obj.modifiers.new("THICKNESS", "SOLIDIFY")
                solidify.thickness = th
                solidify.use_even_offset = True
                solidify.offset = 1
                solidify.show_expanded = False

            # make the shading look good
            esp = obj.modifiers.new("EdgeSplit", "EDGE_SPLIT")
            esp.show_expanded = False

    def _subdivision(self, obj, entity):
        if entity.subdivision_levels > 0:
            subd = obj.modifiers.new("SubD", "SUBSURF")
            subd.levels = entity.subdivision_levels
            subd.show_expanded = False

    def polys_to_mesh(self, entities, scene, name):
        d = bpy.data.meshes.new(name)
        bm = bmesh.new()
        m = Matrix(((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1)))
        for en in entities:
            t = m
            verts = []
            if is_.extrusion(en):
                t = convert.extrusion_to_matrix(en)
            for p in en.points:
                verts.append(bm.verts.new(self.proj((t*Vector(p)).to_3d())))
            if len(verts) > 2:
                bm.faces.new(verts)
            elif len(verts) == 2:
                bm.edges.new(verts)

        bm.to_mesh(d)
        o = bpy.data.objects.new(name, d)
        scene.collection.objects.link(o)
        return o

    def object_mesh(self, entities, scene, name):
        """
        entities: list of DXF entities
        name: name of the returned Blender object (String)
        Accumulates all entities into a Blender bmesh and returns a Blender object containing it.
        """
        d = bpy.data.meshes.new(name)
        bm = bmesh.new()

        i = 0
        for en in entities:
            i += 1
            if en.dxftype == "3DFACE":
                self.the3dface(en, bm)
            else:
                dxftype = getattr(self, en.dxftype.lower(), None)
                if dxftype is not None:
                    dxftype(en, bm)
                else:
                    self.errors.add(en.dxftype.lower() + " - unknown dxftype")
        if i > 0:
            if hasattr(en, "thickness"):
                if en.thickness != 0:
                    self._thickness(bm, en.thickness)
            bm.to_mesh(d)
            o = bpy.data.objects.new(name, d)
            # for POLYFACE
            if hasattr(en, "extrusion"):
                self._extrusion(o, en)
            if hasattr(en, "subdivision_levels"):
                self._subdivision(o, en)
            return o
        return None

    def object_curve(self, entities, scene, name):
        """
        entities: list of DXF entities
        name: name of the returned Blender object (String)
        Accumulates all entities in the list into a Blender curve and returns a Blender object containing it.
        """
        d = bpy.data.curves.new(name, "CURVE")

        i = 0
        lines = []
        for en in entities:
            i += 1
            TYPE = en.dxftype
            if TYPE == "LINE" and self.merge_lines:
                lines.append(en)
                continue
            typefunc = getattr(self, TYPE.lower(), None)
            if typefunc is not None:
                typefunc(en, d)
            else:
                self.errors.add(en.dxftype.lower() + " - unknown dxftype")

        if len(lines) > 0:
            self._merge_lines(lines, d)

        if i > 0:
            self._check3D_object(d)
            o = bpy.data.objects.new(name, d)
            self._thickness_and_width(o, en, scene)
            self._extrusion(o, en)
            return o

        return None

    def object_surface(self, entities, scene, name):
        """
        entities: list of DXF entities
        name: name of the returned Blender object (String) (for future use and also to make it callable from
              _call_types()
        Returns None. Exports all NURB entities to ACIS files if the GUI option for it is set.
        """
        def _get_acis_filename(name, ending):
            df = self.dwg.filename
            dir = os.path.dirname(df)
            filename = bpy.path.display_name(df)
            return os.path.join(dir, "{}_{}.{}".format(filename, name, ending))

        if self.export_acis:
            for en in entities:
                if name in self.acis_files:
                    name = name + "." + str(len([n for n in self.acis_files if name in n])).zfill(3)
                # store SAB files
                if self.dwg.header.get("$ACADVER", "AC1024") > "AC1024":
                    filename = _get_acis_filename(name, "sab")
                    self.acis_files.append(name)
                    with open(filename, 'wb') as f:
                        f.write(en.acis)
                # store SAT files
                else:
                    filename = _get_acis_filename(name, "sat")
                    self.acis_files.append(name)
                    with open(filename, 'w') as f:
                        f.write('\n'.join(en.acis))
        return None

    """ ITERATE OVER DXF ENTITIES AND CREATE BLENDER OBJECTS """

    def _get_group(self, name):
        """
        name: name of group (String)
        Finds group by name or creates it if it does not exist.
        """
        groups = bpy.data.collections
        if name in groups.keys():
            group = groups[name]
        else:
            group = bpy.data.collections.new(name)
        return group

    def _call_object_types(self, TYPE, entities, group, name, scene, separated=False):
        """
        TYPE: DXF type
        entities: list of DXF entities
        group: Blender group (type: bpy_types.Group)
        name: name of the object that is being created and returned (String)
        separated: flag to make _call_types uniformly available for combined_objects() and separated_objects()
        """
        if separated:
            entity = entities[0]
        else:
            entity = entities

        # call merged geometry methods

        if TYPE is True:  # TYPE == True == is_.closed_poly_no_bulge for all entities
            o = self.polys_to_mesh(entities, scene, name)
        elif is_.mesh(TYPE):
            o = self.object_mesh(entities, scene, name)
        elif is_.curve(TYPE):
            o = self.object_curve(entities, scene, name)
        elif is_.nurbs(TYPE):
            o = self.object_surface(entities, scene, name)

        # call separate object methods (or merged geometry if TYPE depending on type)
        else:
            try:
                type_func = getattr(self, TYPE.lower(), None)
                o = type_func(entity, scene, name)
            except (AttributeError, TypeError):
                # don't call self.light(en), self.mtext(o, en), self.text(o, en) with a list of entities
                if is_.separated(TYPE) and not separated:
                    self.errors.add("DXF-Import: multiple %ss cannot be merged into a Blender object." % TYPE)
                elif TYPE == "not_mergeable":
                    self.errors.add("DXF-Import: Not mergeable dxf type '%s' should not be called in merge-mode" % TYPE)
                else:
                    self.errors.add("DXF-import: Unsupported dxftype: %s" % TYPE)
                raise

        if type(o) == bpy.types.Object:
            if o.name not in scene.objects:
                scene.collection.objects.link(o)

            if o.name not in group.objects:
                group.objects.link(o)
        return o

    def _recenter(self, scene, name):
        bpy.context.screen.scene = scene
        bpy.context.view_layer.update()
        bpy.ops.object.select_all(action='DESELECT')

        recenter_objects = (o for o in scene.objects if "BEVEL" not in o.name and "TAPER" not in o.name
                            and o not in self.objects_before)
        xmin, ymin, zmin, xmax, ymax, zmax = self._bbox(recenter_objects)
        vmin = Vector((xmin, ymin, zmin))
        vmax = Vector((xmax, ymax, zmax))
        center = vmin + (vmax - vmin) / 2
        for o in (o for o in scene.objects if "BEVEL" not in o.name and "TAPER" not in o.name
                  and o not in self.objects_before and o.parent is None):
            o.location = o.location - center
            o.select_set(True)

        if not self.did_group_instance:
            bpy.ops.object.origin_set(type='ORIGIN_CURSOR')

        if self.pDXF is not None:
            self.georeference(scene, center)
        else:
            scene[name + "_recenter"] = center

    def _dupliface(self, blockname, inserts, blgroup, scene):
        """
        go through all inserts and check if there is one having no rotation or extrusion (if there is none
        then there is none... any other measure to hide the original is also not water proof; just try the first
        and most obvious way and if it doesn't work, we will not try to cover uncoverable cases)
        - Place the duplicator object with a first face according to the chosen start insert.
        - Add the block as a child object to the duplicator object. If there are any recursive inserts in the
          insert, keep appending children to the children.
        - Any subsequent inserts in the list are represented just by a face in the duplicator object.
        """
        aunits = self.dwg.header.get('$AUNITS', 0)
        f = 20
        base = [
            Vector(( sqrt(((1/f)**2))/2, sqrt(((1/f)**2))/2,0)),
            Vector(( sqrt(((1/f)**2))/2,-sqrt(((1/f)**2))/2,0)),
            Vector((-sqrt(((1/f)**2))/2,-sqrt(((1/f)**2))/2,0)),
            Vector((-sqrt(((1/f)**2))/2, sqrt(((1/f)**2))/2,0)),
        ]

        bm = bmesh.new()
        location = None
        for entity in inserts:
            extrusion = convert.extrusion_to_matrix(entity)
            scale = Matrix(((entity.scale[0],0,0,0),(0,entity.scale[1],0,0),(0,0,entity.scale[2],0),(0,0,0,1)))
            rotation = radians(entity.rotation) if aunits == 0 else entity.rotation
            rotm = scale @ extrusion @ extrusion.Rotation(rotation, 4, "Z")
            if location is None:
                location = rotm @ Vector(entity.insert)
                entity.insert = (0, 0, 0)
                transformation = rotm
            else:
                transformation = rotm.Translation((extrusion @ Vector(entity.insert))-location) @ rotm
            verts = []
            for v in base:
                verts.append(bm.verts.new(transformation @ v))
            bm.faces.new(verts)



        m = bpy.data.meshes.new(blockname+"_geometry")
        bm.to_mesh(m)
        o = bpy.data.objects.new(blockname, m)
        o.location = location
        scene.collection.objects.link(o)

        self._nest_block(o, blockname, blgroup, scene)
        o.instance_type = "FACES"
        o.use_instance_faces_scale = True
        o.instance_faces_scale = f

    def _nest_block(self, parent, name, blgroup, scene):
        b = self.dwg.blocks[name]
        e = bpy.data.objects.new(name, None)
        scene.collection.objects.link(e)
        #e.location = parent.location
        e.parent = parent
        for TYPE, grouped in groupsort.by_dxftype(b):
            if TYPE == "INSERT":
                for en in grouped:
                    self._nest_block(e, en.name, blgroup, scene)
            else:
                o = self._call_object_types(TYPE, grouped, blgroup, name+"_"+TYPE, scene)
                #o.location = e.location
                o.parent = e

    def combined_objects(self, entities, scene, override_name=None, override_group=None):
        """
        entities: list of dxf entities
        override_group & override_name: for use within insert() and block()
        Adds multiple dxf entities to one Blender object (per blender or dxf type).
        """
        objects = []
        for layer_name, layer_ents in groupsort.by_layer(entities):
            # group and name
            if override_group is None:
                group = self._get_group(layer_name)
            else:
                group = override_group
            if override_name is not None:
                layer_name = override_name

            # sort
            if self.combination == BY_LAYER:
                group_sorted = groupsort.by_blender_type(layer_ents)
            elif self.combination == BY_DXFTYPE or self.combination == BY_BLOCK:
                group_sorted = groupsort.by_dxftype(layer_ents)
            elif self.combination == BY_CLOSED_NO_BULGE_POLY:
                group_sorted = groupsort.by_closed_poly_no_bulge(layer_ents)
            else:
                break

            for TYPE, grouped_entities in group_sorted:
                if self.but_group_by_att and self.combination != BY_CLOSED_NO_BULGE_POLY and self.combination != BY_BLOCK:
                    for atts, by_att in groupsort.by_attributes(grouped_entities):
                        thickness, subd, width, extrusion = atts
                        if extrusion is None:  # unset extrusion defaults to (0, 0, 1)
                            extrusion = (0, 0, 1)
                        att = ""
                        if thickness != 0:
                            att += "thickness" + str(thickness) + ", "
                        if subd > 0:
                            att += "subd" + str(subd) + ", "
                        if width != [(0, 0)]:
                            att += "width" + str(width) + ", "
                        if extrusion != (0, 0, 1):
                            att += "extrusion" + str([str(round(c, 1)) + ".." + str(c)[-1:] for c in extrusion]) + ", "
                        name = layer_name + "_" + TYPE.replace("object_", "") + "_" + att

                        o = self._call_object_types(TYPE, by_att, group, name, scene, False)
                        if o is not None:
                            objects.append(o)
                else:
                    if type(TYPE) is bool and not TYPE:
                        for ttype, sub_entities in groupsort.by_blender_type(grouped_entities):
                            name = layer_name + "_" + ttype.replace("object_", "")
                            o = self._call_object_types(ttype, sub_entities, group, name, scene, False)
                            if o is not None:
                                objects.append(o)
                    else:
                        if TYPE == "INSERT" and self.combination == BY_BLOCK:
                            for NAME, grouped_inserts in groupsort.by_insert_block_name(grouped_entities):
                                sorted_inserts = []
                                separates = []
                                for i in grouped_inserts:
                                    sames = 1
                                    for c in range(2):
                                        if i.scale[c+1] - i.scale[0] < 0.00001:
                                            sames += 1
                                    if not (sames == 3 or (sames == 2 and i.scale[2] == 1)):
                                        print(i.scale)
                                        separates.append(i)
                                    else:
                                        if i.extrusion == (0, 0, 1) and i.rotation == 0.0 and i.scale == (1, 1, 1):
                                            sorted_inserts.insert(0, i)
                                        else:
                                            sorted_inserts.append(i)

                                if len(sorted_inserts) > 0:
                                    self._dupliface(NAME, sorted_inserts, group, scene)
                                for s in separates:
                                    self.insert(s, scene, NAME, group)
                        else:
                            name = layer_name + "_" + TYPE.replace("object_", "") if type(TYPE) is str else "MERGED_POLYS"
                            o = self._call_object_types(TYPE, grouped_entities, group, name, scene, False)
                            if o is not None:
                                objects.append(o)
        return objects

    def separated_entities(self, entities, scene, override_name=None, override_group=None):
        """
        entities: list of dxf entities
        override_group & override_name: for use within insert() and block()
        Adds multiple DXF entities to multiple Blender objects.
        """
        def _do_it(en):
            # group and name
            if override_group is None:
                group = self._get_group(en.layer)
            else:
                group = override_group

            if override_name is None:
                name = en.dxftype
            else:
                name = override_name

            if en.dxftype == "POINT":
                o = self.point_object(en)
            else:
                o = self._call_object_types(en.dxftype, [en], group, name, scene, separated=True)

            if o is not None:
                objects.append(o)

        objects = []
        split_entities = []

        for en in entities:
            split = convert.split_by_width(en)
            if len(split) > 1:
                split_entities.extend(split)
                continue
            _do_it(en)

        for en in split_entities:
            _do_it(en)

        return objects

    def entities(self, name, scene=None):
        """
        Iterates over all DXF entities according to the options set by user.
        """
        if scene is None:
            scene = bpy.context.scene

        self.current_scene = scene

        if self.recenter:
            self.objects_before += scene.objects[:]

        if self.combination == BY_BLOCK:
            self.combined_objects((en for en in self.dwg.modelspace()), scene)
        elif self.combination != SEPARATED:
            self.combined_objects((en for en in self.dwg.modelspace() if is_.combined_entity(en)), scene)
            self.separated_entities((en for en in self.dwg.modelspace() if is_.separated_entity(en)), scene)
        else:
            self.separated_entities((en for en in self.dwg.modelspace() if en.dxftype != "ATTDEF"), scene)

        if self.recenter:
            self._recenter(scene, name)
        elif self.pDXF is not None:
            self.georeference(scene, Vector((0, 0, 0)))

        if type(self.pScene) is TransverseMercator:
            scene['SRID'] = "tmerc"
        elif self.pScene is not None:  # assume Proj
            scene['SRID'] = re.findall("\+init=(.+)\s", self.pScene.srs)[0]

        #bpy.context.screen.scene = scene

        return self.errors
        # trying to import dimensions:
        # self.separated_objects((block for block in self.dwg.blocks if block.name.startswith("*")))



# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------
import bpy
import os
import subprocess
from bl_operators.presets import AddPresetBase
from mathutils import Vector
from bpy.props import StringProperty
from .archipack_gl import (
    ThumbHandle, Screen, GlRect,
    GlPolyline, GlPolygon, GlText, GlHandle
)
preset_paths = bpy.utils.script_paths("presets")
addons_paths = bpy.utils.script_paths("addons")


class CruxHandle(GlHandle):

    def __init__(self, sensor_size, depth):
        GlHandle.__init__(self, sensor_size, 0, True, False)
        self.branch_0 = GlPolygon((1, 1, 1, 1), d=2)
        self.branch_1 = GlPolygon((1, 1, 1, 1), d=2)
        self.branch_2 = GlPolygon((1, 1, 1, 1), d=2)
        self.branch_3 = GlPolygon((1, 1, 1, 1), d=2)
        self.depth = depth

    def set_pos(self, pos_2d):
        self.pos_2d = pos_2d
        o = pos_2d
        w = 0.5 * self.sensor_width
        d = self.depth
        c = d / 1.4242
        s = w - c
        p0 = o + Vector((s, w))
        p1 = o + Vector((w, s))
        p2 = o + Vector((c, 0))
        p3 = o + Vector((w, -s))
        p4 = o + Vector((s, -w))
        p5 = o + Vector((0, -c))
        p6 = o + Vector((-s, -w))
        p7 = o + Vector((-w, -s))
        p8 = o + Vector((-c, 0))
        p9 = o + Vector((-w, s))
        p10 = o + Vector((-s, w))
        p11 = o + Vector((0, c))
        self.branch_0.set_pos([p11, p0, p1, p2, o])
        self.branch_1.set_pos([p2, p3, p4, p5, o])
        self.branch_2.set_pos([p5, p6, p7, p8, o])
        self.branch_3.set_pos([p8, p9, p10, p11, o])

    @property
    def pts(self):
        return [self.pos_2d]

    @property
    def sensor_center(self):
        return self.pos_2d

    def draw(self, context, render=False):
        self.render = render
        self.branch_0.colour_inactive = self.colour
        self.branch_1.colour_inactive = self.colour
        self.branch_2.colour_inactive = self.colour
        self.branch_3.colour_inactive = self.colour
        self.branch_0.draw(context)
        self.branch_1.draw(context)
        self.branch_2.draw(context)
        self.branch_3.draw(context)


class SeekBox(GlText, GlHandle):
    """
        Text input to filter items by label
        TODO:
            - add cross to empty text
            - get text from keyboard
    """

    def __init__(self):
        GlHandle.__init__(self, 0, 0, True, False, d=2)
        GlText.__init__(self, d=2)
        self.sensor_width = 250
        self.pos_3d = Vector((0, 0))
        self.bg = GlRect(colour=(0, 0, 0, 0.7))
        self.frame = GlPolyline((1, 1, 1, 1), d=2)
        self.frame.closed = True
        self.cancel = CruxHandle(16, 4)
        self.line_pos = 0

    @property
    def pts(self):
        return [self.pos_3d]

    def set_pos(self, context, pos_2d):
        x, ty = self.text_size(context)
        w = self.sensor_width
        y = 12
        pos_2d.y += y
        pos_2d.x -= 0.5 * w
        self.pos_2d = pos_2d.copy()
        self.pos_3d = pos_2d.copy()
        self.pos_3d.x += 6
        self.sensor_height = y
        p0 = pos_2d + Vector((w, -0.5 * y))
        p1 = pos_2d + Vector((w, 1.5 * y))
        p2 = pos_2d + Vector((0, 1.5 * y))
        p3 = pos_2d + Vector((0, -0.5 * y))
        self.bg.set_pos([p0, p2])
        self.frame.set_pos([p0, p1, p2, p3])
        self.cancel.set_pos(pos_2d + Vector((w + 15, 0.5 * y)))

    def keyboard_entry(self, context, event):
        c = event.ascii
        if c:
            if c == ",":
                c = "."
            self.label = self.label[:self.line_pos] + c + self.label[self.line_pos:]
            self.line_pos += 1

        if self.label:
            if event.type == 'BACK_SPACE':
                self.label = self.label[:self.line_pos - 1] + self.label[self.line_pos:]
                self.line_pos -= 1

            elif event.type == 'DEL':
                self.label = self.label[:self.line_pos] + self.label[self.line_pos + 1:]

            elif event.type == 'LEFT_ARROW':
                self.line_pos = (self.line_pos - 1) % (len(self.label) + 1)

            elif event.type == 'RIGHT_ARROW':
                self.line_pos = (self.line_pos + 1) % (len(self.label) + 1)

    def draw(self, context):
        self.bg.draw(context)
        self.frame.draw(context)
        GlText.draw(self, context)
        self.cancel.draw(context)

    @property
    def sensor_center(self):
        return self.pos_3d


class PresetMenuItem():
    def __init__(self, thumbsize, preset, image=None):
        name = bpy.path.display_name_from_filepath(preset)
        self.preset = preset
        self.image = image
        self.image.gl_load()
        self.handle = ThumbHandle(thumbsize, name, self.image, draggable=True)
        self.enable = True

    def filter(self, keywords):
        for key in keywords:
            if key not in self.handle.label.label:
                return False
        return True

    def cleanup(self):
        if self.image:
            self.image.gl_free()
            # bpy.data.images.remove(self.image)

    def set_pos(self, context, pos):
        self.handle.set_pos(context, pos)

    def check_hover(self, mouse_pos):
        self.handle.check_hover(mouse_pos)

    def mouse_press(self):
        if self.handle.hover:
            self.handle.hover = False
            self.handle.active = True
            return True
        return False

    def draw(self, context):
        if self.enable:
            self.handle.draw(context)


class PresetMenu():

    keyboard_type = {
            'BACK_SPACE', 'DEL',
            'LEFT_ARROW', 'RIGHT_ARROW'
            }

    def __init__(self, context, category, thumbsize=Vector((150, 100))):
        self.imageList = []
        self.menuItems = []
        self.thumbsize = thumbsize
        file_list = self.scan_files(category)
        self.default_image = None
        self.load_default_image()
        for filepath in file_list:
            self.make_menuitem(filepath)
        self.margin = 50
        self.y_scroll = 0
        self.scroll_max = 1000
        self.spacing = Vector((25, 25))
        self.screen = Screen(self.margin)
        self.mouse_pos = Vector((0, 0))
        self.bg = GlRect(colour=(0, 0, 0, 0.7))
        self.border = GlPolyline((0.7, 0.7, 0.7, 1), d=2)
        self.keywords = SeekBox()
        self.keywords.colour_normal = (1, 1, 1, 1)
        self.border.closed = True
        self.set_pos(context)

    def load_default_image(self):
        img_idx = bpy.data.images.find("missing.png")
        if img_idx > -1:
            self.default_image = bpy.data.images[img_idx]
            self.imageList.append(self.default_image.filepath_raw)
            return
        dir_path = os.path.dirname(os.path.realpath(__file__))
        sub_path = "presets" + os.path.sep + "missing.png"
        filepath = os.path.join(dir_path, sub_path)
        if os.path.exists(filepath) and os.path.isfile(filepath):
            self.default_image = bpy.data.images.load(filepath=filepath)
            self.imageList.append(self.default_image.filepath_raw)
        if self.default_image is None:
            raise EnvironmentError("archipack/presets/missing.png not found")

    def scan_files(self, category):
        file_list = []
        """
        # load default presets
        dir_path = os.path.dirname(os.path.realpath(__file__))
        sub_path = "presets" + os.path.sep + category
        presets_path = os.path.join(dir_path, sub_path)
        if os.path.exists(presets_path):
            file_list += [presets_path + os.path.sep + f[:-3]
                for f in os.listdir(presets_path)
                if f.endswith('.py') and
                not f.startswith('.')]
        """
        # load user def presets
        for path in preset_paths:
            presets_path = os.path.join(path, category)
            if os.path.exists(presets_path):
                file_list += [presets_path + os.path.sep + f[:-3]
                    for f in os.listdir(presets_path)
                    if f.endswith('.py') and
                    not f.startswith('.')]

        file_list.sort()
        return file_list

    def clearImages(self):
        for item in self.menuItems:
            item.cleanup()
        for image in bpy.data.images:
            if image.filepath_raw in self.imageList:
                # image.user_clear()
                bpy.data.images.remove(image, do_unlink=True)
        self.imageList.clear()


    def make_menuitem(self, filepath):
        """
            @TODO:
            Lazy load images
        """
        image = None
        img_idx = bpy.data.images.find(os.path.basename(filepath) + '.png')
        if img_idx > -1 and bpy.data.images[img_idx].filepath_raw == filepath:
            image = bpy.data.images[img_idx]
            self.imageList.append(image.filepath_raw)
        elif os.path.exists(filepath + '.png') and os.path.isfile(filepath + '.png'):
            image = bpy.data.images.load(filepath=filepath + '.png')
            if hasattr(image, "colorspace_settings"):
                image.colorspace_settings.name = 'Raw'
            self.imageList.append(image)
        if image is None:
            image = self.default_image
        item = PresetMenuItem(self.thumbsize, filepath + '.py', image)
        self.menuItems.append(item)

    def set_pos(self, context):

        x_min, x_max, y_min, y_max = self.screen.size(context)
        y_max -= 20
        p0, p1, p2, p3 = Vector((x_min, y_min)), Vector((x_min, y_max)), Vector((x_max, y_max)), Vector((x_max, y_min))
        self.bg.set_pos([p0, p2])
        self.border.set_pos([p0, p1, p2, p3])
        x_min += 0.5 * self.thumbsize.x + 0.5 * self.margin
        x_max -= 0.5 * self.thumbsize.x + 0.5 * self.margin
        y_max -= 0.5 * self.thumbsize.y + 0.5 * self.margin
        y_min += 0.5 * self.margin
        x = x_min
        y = y_max + self.y_scroll
        n_rows = 0

        self.keywords.set_pos(context, p1 + 0.5 * (p2 - p1))
        keywords = self.keywords.label.split(" ")

        for item in self.menuItems:
            if y > y_max or y < y_min:
                item.enable = False
            else:
                item.enable = True

            # filter items by name
            if len(keywords) > 0 and not item.filter(keywords):
                item.enable = False
                continue

            item.set_pos(context, Vector((x, y)))
            x += self.thumbsize.x + self.spacing.x
            if x > x_max:
                n_rows += 1
                x = x_min
                y -= self.thumbsize.y + self.spacing.y

        self.scroll_max = max(0, n_rows - 1) * (self.thumbsize.y + self.spacing.y)

    def draw(self, context):
        self.bg.draw(context)
        self.border.draw(context)
        self.keywords.draw(context)
        for item in self.menuItems:
            item.draw(context)

    def mouse_press(self, context, event):
        self.mouse_position(event)

        if self.keywords.cancel.hover:
            self.keywords.label = ""
            self.keywords.line_pos = 0
            self.set_pos(context)

        for item in self.menuItems:
            if item.enable and item.mouse_press():
                # load item preset
                return item.preset
        return None

    def mouse_position(self, event):
        self.mouse_pos.x, self.mouse_pos.y = event.mouse_region_x, event.mouse_region_y

    def mouse_move(self, context, event):
        self.mouse_position(event)
        self.keywords.check_hover(self.mouse_pos)
        self.keywords.cancel.check_hover(self.mouse_pos)
        for item in self.menuItems:
            item.check_hover(self.mouse_pos)

    def scroll_up(self, context, event):
        self.y_scroll = max(0, self.y_scroll - (self.thumbsize.y + self.spacing.y))
        self.set_pos(context)
        # print("scroll_up %s" % (self.y_scroll))

    def scroll_down(self, context, event):
        self.y_scroll = min(self.scroll_max, self.y_scroll + (self.thumbsize.y + self.spacing.y))
        self.set_pos(context)
        # print("scroll_down %s" % (self.y_scroll))

    def keyboard_entry(self, context, event):
        self.keywords.keyboard_entry(context, event)
        self.set_pos(context)


class PresetMenuOperator():

    preset_operator : StringProperty(
        options={'SKIP_SAVE'},
        default="script.execute_preset"
    )

    def __init__(self):
        self.menu = None
        self._handle = None

    def exit(self, context):
        self.menu.clearImages()
        bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

    def draw_handler(self, _self, context):
        self.menu.draw(context)

    def modal(self, context, event):
        if self.menu is None:
            return {'FINISHED'}
        context.area.tag_redraw()
        if event.type == 'MOUSEMOVE':
            self.menu.mouse_move(context, event)
        elif event.type == 'WHEELUPMOUSE' or \
                (event.type == 'UP_ARROW' and event.value == 'PRESS'):
            self.menu.scroll_up(context, event)
        elif event.type == 'WHEELDOWNMOUSE' or \
                (event.type == 'DOWN_ARROW' and event.value == 'PRESS'):
            self.menu.scroll_down(context, event)
        elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':
            preset = self.menu.mouse_press(context, event)
            if preset is not None:
                self.exit(context)
                po = self.preset_operator.split(".")
                op = getattr(getattr(bpy.ops, po[0]), po[1])
                if self.preset_operator == 'script.execute_preset':
                    # call from preset menu
                    # ensure right active_object class
                    o = context.active_object
                    if o.data and self.preset_subdir in o.data:
                        d = getattr(o.data, self.preset_subdir)[0]
                    elif self.preset_subdir in o:
                        d = getattr(o, self.preset_subdir)[0]
                    if d is not None:
                        d.auto_update = False
                        # print("Archipack execute_preset loading auto_update:%s" % d.auto_update)
                        op('INVOKE_DEFAULT', filepath=preset, menu_idname=self.bl_idname)
                        # print("Archipack execute_preset loaded  auto_update: %s" % d.auto_update)
                        d.auto_update = True
                else:
                    # call draw operator
                    if op.poll():
                        op('INVOKE_DEFAULT', filepath=preset)
                    else:
                        print("Poll failed")
                return {'FINISHED'}
        elif event.ascii or (
                event.type in self.menu.keyboard_type and
                event.value == 'RELEASE'):
            self.menu.keyboard_entry(context, event)
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            self.exit(context)
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.area.type == 'VIEW_3D':

            # with alt pressed on invoke, will bypass menu operator and
            # call preset_operator
            # allow start drawing linked copy of active object
            if event.alt or event.ctrl:
                po = self.preset_operator.split(".")
                op = getattr(getattr(bpy.ops, po[0]), po[1])
                d = context.active_object.data

                if d is not None and self.preset_subdir in d and op.poll():
                    op('INVOKE_DEFAULT')
                else:
                    self.report({'WARNING'}, "Active object must be a " + self.preset_subdir.split("_")[1].capitalize())
                    return {'CANCELLED'}
                return {'FINISHED'}

            self.menu = PresetMenu(context, self.preset_subdir)

            # the arguments we pass the the callback
            args = (self, context)
            # Add the region OpenGL drawing callback
            # draw in view space with 'POST_VIEW' and 'PRE_VIEW'
            self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_handler, args, 'WINDOW', 'POST_PIXEL')
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot show preset flinger")
            return {'CANCELLED'}


class ArchipackPreset(AddPresetBase):

    @classmethod
    def poll(cls, context):
        o = context.active_object
        return o is not None and \
            o.data is not None and \
            "archipack_" + cls.__name__[13:-7] in o.data

    @property
    def preset_subdir(self):
        return "archipack_" + self.__class__.__name__[13:-7]

    @property
    def blacklist(self):
        """
            properties black list for presets
            may override on addon basis
        """
        return []

    @property
    def preset_values(self):
        blacklist = self.blacklist
        blacklist.extend(bpy.types.Mesh.bl_rna.properties.keys())
        d = getattr(bpy.context.active_object.data, self.preset_subdir)[0]
        props = d.rna_type.bl_rna.properties.items()
        ret = []
        for prop_id, prop in props:
            if prop_id not in blacklist:
                if not (prop.is_hidden or prop.is_skip_save):
                    ret.append("d.%s" % prop_id)
        ret.sort()
        return ret

    @property
    def preset_defines(self):
        o = bpy.context.active_object
        m = o.archipack_material[0]
        return [
            "d = bpy.context.active_object.data." + self.preset_subdir + "[0]",
            "bpy.ops.archipack.material(category='" + m.category + "', material='" + m.material + "')"
        ]

    def pre_cb(self, context):
        return

    def remove(self, context, filepath):
        # remove preset
        os.remove(filepath)
        # remove thumb
        os.remove(filepath[:-3] + ".png")

    def background_render(self, context, cls, preset):
        generator = os.path.dirname(os.path.realpath(__file__)) + os.path.sep + "archipack_thumbs.py"
        addon_name = __name__.split('.')[0]
        matlib_path = context.preferences.addons[addon_name].preferences.matlib_path
        # Run external instance of blender like the original thumbnail generator.
        cmd = [
            bpy.app.binary_path,
            "--background",
            "--factory-startup",
            "-noaudio",
            # "--addons", addon_name,
            "--python", generator,
            "--",
            "addon:" + addon_name,
            "matlib:" + matlib_path,
            "cls:" + cls,
            "preset:" + preset
        ]

        subprocess.Popen(cmd)

    def post_cb(self, context):

        if not self.remove_active:

            name = self.name.strip()
            if not name:
                return

            filename = self.as_filename(name)
            target_path = os.path.join("presets", self.preset_subdir)
            target_path = bpy.utils.user_resource('SCRIPTS',
                                                  target_path,
                                                  create=True)

            preset = os.path.join(target_path, filename) + ".py"
            cls = self.preset_subdir[10:]
            # print("post cb cls:%s preset:%s" % (cls, preset))
            self.background_render(context, cls, preset)

            return



#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
Material Selector

Quickly switch materials in the active mesh without going to the Properties editor

Based on 'Afeitadora's work on Elysiun
http://www.elysiun.com/forum/showthread.php?290097-Dynamic-Object-Dropdown-List&p=2361851#post2361851

"""

import bpy

def ui_node_editor_material_select(self, context):

    act_ob = context.active_object

    if act_ob and context.active_object.type in {'MESH', 'CURVE', 'SURFACE', 'META'} and \
        context.space_data.tree_type == 'ShaderNodeTree' and \
        context.space_data.shader_type == 'OBJECT':

        if act_ob.active_material:
            mat_name = act_ob.active_material.name
        else:
            mat_name = "No Material"

        self.layout.operator_menu_enum("material.menu_select",
                                       "material_select",
                                        text=mat_name,
                                       icon="MATERIAL")

class AMNodeEditorMaterialSelect(bpy.types.Operator):
    bl_idname = "material.menu_select"
    bl_label = "Select Material"
    bl_description = "Switch to another material in this mesh"

    def avail_materials(self,context):
        items = [(str(i),x.name,x.name, "MATERIAL", i) for i,x in enumerate(bpy.context.active_object.material_slots)]
        return items
    material_select: bpy.props.EnumProperty(items = avail_materials, name = "Available Materials")

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self,context):
        bpy.context.active_object.active_material_index = int(self.material_select)
        return {'FINISHED'}

def register():
    bpy.utils.register_class(AMNodeEditorMaterialSelect)
    bpy.types.NODE_HT_header.append(ui_node_editor_material_select)

def unregister():
    bpy.utils.unregister_class(AMNodeEditorMaterialSelect)
    bpy.types.NODE_HT_header.remove(ui_node_editor_material_select)



#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================

# <pep8 compliant>

import bpy
import math
import inspect
import functools

from mathutils import Matrix

from .errors import MetarigError
from .collections import ensure_widget_collection

WGT_PREFIX = "WGT-"  # Prefix for widget objects

#=============================================
# Widget creation
#=============================================


def obj_to_bone(obj, rig, bone_name, bone_transform_name=None):
    """ Places an object at the location/rotation/scale of the given bone.
    """
    if bpy.context.mode == 'EDIT_ARMATURE':
        raise MetarigError("obj_to_bone(): does not work while in edit mode")

    bone = rig.pose.bones[bone_name]
    scale = bone.custom_shape_scale

    if bone.use_custom_shape_bone_size:
        scale *= bone.length

    if bone_transform_name is not None:
        bone = rig.pose.bones[bone_transform_name]
    elif bone.custom_shape_transform:
        bone = bone.custom_shape_transform

    obj.rotation_mode = 'XYZ'
    obj.matrix_basis = rig.matrix_world @ bone.bone.matrix_local @ Matrix.Scale(scale, 4)


def create_widget(rig, bone_name, bone_transform_name=None, *, widget_name=None, widget_force_new=False):
    """ Creates an empty widget object for a bone, and returns the object.
    """
    assert rig.mode != 'EDIT'

    obj_name = widget_name or WGT_PREFIX + rig.name + '_' + bone_name
    scene = bpy.context.scene
    collection = ensure_widget_collection(bpy.context, 'WGTS_' + rig.name)
    reuse_mesh = None

    # Check if it already exists in the scene
    if not widget_force_new:
        if obj_name in scene.objects:
            # Move object to bone position, in case it changed
            obj = scene.objects[obj_name]
            obj_to_bone(obj, rig, bone_name, bone_transform_name)

            return None

        # Delete object if it exists in blend data but not scene data.
        # This is necessary so we can then create the object without
        # name conflicts.
        if obj_name in bpy.data.objects:
            bpy.data.objects.remove(bpy.data.objects[obj_name])

        # Create a linked duplicate of the widget assigned in the metarig
        reuse_widget = rig.pose.bones[bone_name].custom_shape
        if reuse_widget:
            reuse_mesh = reuse_widget.data

    # Create mesh object
    mesh = reuse_mesh or bpy.data.meshes.new(obj_name)
    obj = bpy.data.objects.new(obj_name, mesh)
    collection.objects.link(obj)

    # Move object to bone position and set layers
    obj_to_bone(obj, rig, bone_name, bone_transform_name)

    if reuse_mesh:
        return None

    return obj


#=============================================
# Widget choice dropdown
#=============================================

_registered_widgets = {}


def _get_valid_args(callback, skip):
    spec = inspect.getfullargspec(callback)
    return set(spec.args[skip:] + spec.kwonlyargs)


def register_widget(name, callback, **default_args):
    unwrapped = inspect.unwrap(callback)
    if unwrapped != callback:
        valid_args = _get_valid_args(unwrapped, 1)
    else:
        valid_args = _get_valid_args(callback, 2)

    _registered_widgets[name] = (callback, valid_args, default_args)


def layout_widget_dropdown(layout, props, prop_name, **kwargs):
    "Create a UI dropdown to select a widget from the known list."

    id_store = bpy.context.window_manager
    rigify_widgets = id_store.rigify_widgets

    rigify_widgets.clear()

    for name in sorted(_registered_widgets):
        item = rigify_widgets.add()
        item.name = name

    layout.prop_search(props, prop_name, id_store, "rigify_widgets", **kwargs)


def create_registered_widget(obj, bone_name, widget_id, **kwargs):
    try:
        callback, valid_args, default_args = _registered_widgets[widget_id]
    except KeyError:
        raise MetarigError("Unknown widget name: " + widget_id)

    # Convert between radius and size
    if kwargs.get('size') and 'size' not in valid_args:
        if 'radius' in valid_args and not kwargs.get('radius'):
            kwargs['radius'] = kwargs['size'] / 2

    elif kwargs.get('radius') and 'radius' not in valid_args:
        if 'size' in valid_args and not kwargs.get('size'):
            kwargs['size'] = kwargs['radius'] * 2

    args = { **default_args, **kwargs }

    return callback(obj, bone_name, **{ k:v for k,v in args.items() if k in valid_args})


#=============================================
# Widget geometry
#=============================================

class GeometryData:
    def __init__(self):
        self.verts = []
        self.edges = []
        self.faces = []


def widget_generator(generate_func=None, *, register=None, subsurf=0):
    if generate_func is None:
        return functools.partial(widget_generator, register=register, subsurf=subsurf)

    """
    Decorator that encapsulates a call to create_widget, and only requires
    the actual function to fill the provided vertex and edge lists.

    Accepts parameters of create_widget, plus any keyword arguments the
    wrapped function has.
    """
    @functools.wraps(generate_func)
    def wrapper(rig, bone_name, bone_transform_name=None, widget_name=None, widget_force_new=False, **kwargs):
        obj = create_widget(rig, bone_name, bone_transform_name, widget_name=widget_name, widget_force_new=widget_force_new)
        if obj is not None:
            geom = GeometryData()

            generate_func(geom, **kwargs)

            mesh = obj.data
            mesh.from_pydata(geom.verts, geom.edges, geom.faces)
            mesh.update()

            if subsurf:
                mod = obj.modifiers.new("subsurf", 'SUBSURF')
                mod.levels = subsurf

            return obj
        else:
            return None

    if register:
        register_widget(register, wrapper)

    return wrapper


def create_circle_polygon(number_verts, axis, radius=1.0, head_tail=0.0):
    """ Creates a basic circle around of an axis selected.
        number_verts: number of vertices of the polygon
        axis: axis normal to the circle
        radius: the radius of the circle
        head_tail: where along the length of the bone the circle is (0.0=head, 1.0=tail)
    """
    verts = []
    edges = []
    angle = 2 * math.pi / number_verts
    i = 0

    assert(axis in 'XYZ')

    while i < (number_verts):
        a = math.cos(i * angle)
        b = math.sin(i * angle)

        if axis == 'X':
            verts.append((head_tail, a * radius, b * radius))
        elif axis == 'Y':
            verts.append((a * radius, head_tail, b * radius))
        elif axis == 'Z':
            verts.append((a * radius, b * radius, head_tail))

        if i < (number_verts - 1):
            edges.append((i , i + 1))

        i += 1

    edges.append((0, number_verts - 1))

    return verts, edges


def adjust_widget_axis(obj, axis='y', offset=0.0):
    mesh = obj.data

    if axis[0] == '-':
        s = -1.0
        axis = axis[1]
    else:
        s = 1.0

    trans_matrix = Matrix.Translation((0.0, offset, 0.0))
    rot_matrix = Matrix.Diagonal((1.0, s, 1.0, 1.0))

    if axis == "x":
        rot_matrix = Matrix.Rotation(-s*math.pi/2, 4, 'Z')
        trans_matrix = Matrix.Translation((offset, 0.0, 0.0))

    elif axis == "z":
        rot_matrix = Matrix.Rotation(s*math.pi/2, 4, 'X')
        trans_matrix = Matrix.Translation((0.0, 0.0, offset))

    matrix = trans_matrix @ rot_matrix

    for vert in mesh.vertices:
        vert.co = matrix @ vert.co


def adjust_widget_transform_mesh(obj, matrix, local=None):
    """Adjust the generated widget by applying a correction matrix to the mesh.
       If local is false, the matrix is in world space.
       If local is True, it's in the local space of the widget.
       If local is a bone, it's in the local space of the bone.
    """
    if obj:
        if local is not True:
            if local:
                assert isinstance(local, bpy.types.PoseBone)
                bonemat = local.id_data.matrix_world @ local.bone.matrix_local
                matrix = bonemat @ matrix @ bonemat.inverted()

            obmat = obj.matrix_basis
            matrix = obmat.inverted() @ matrix @ obmat

        obj.data.transform(matrix)


def write_widget(obj, name='thing', use_size=True):
    """ Write a mesh object as a python script for widget use.
    """
    script = ""
    script += "@widget_generator\n"
    script += "def create_"+name+"_widget(geom";
    if use_size:
        script += ", *, size=1.0"
    script += "):\n"

    # Vertices
    szs = "*size" if use_size else ""
    width = 2 if use_size else 3

    script += "    geom.verts = ["
    for i, v in enumerate(obj.data.vertices):
        script += "({:g}{}, {:g}{}, {:g}{}),".format(v.co[0], szs, v.co[1], szs, v.co[2], szs)
        script += "\n                  " if i % width == (width - 1) else " "
    script += "]\n"

    # Edges
    script += "    geom.edges = ["
    for i, e in enumerate(obj.data.edges):
        script += "(" + str(e.vertices[0]) + ", " + str(e.vertices[1]) + "),"
        script += "\n                  " if i % 10 == 9 else " "
    script += "]\n"

    # Faces
    if obj.data.polygons:
        script += "    geom.faces = ["
        for i, f in enumerate(obj.data.polygons):
            script += "(" + ", ".join(str(v) for v in f.vertices) + "),"
            script += "\n                  " if i % 10 == 9 else " "
        script += "]\n"

    return script



"""
Extending Menus
+++++++++++++++

When creating menus for add-ons you can't reference menus
in Blender's default scripts.
Instead, the add-on can add menu items to existing menus.

The function menu_draw acts like :class:`Menu.draw`.
"""
import bpy


def menu_draw(self, context):
    self.layout.operator("wm.save_homefile")


bpy.types.TOPBAR_MT_file.append(menu_draw)



# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------


class Keymaps:
    """
        Expose user defined keymaps as event
        so in modal operator we are able to
        identify like
        if (event == keymap.undo.event):

        and in feedback panels:
            keymap.undo.key
            keymap.undo.name
    """
    def __init__(self, context):
        """
            Init keymaps properties
        """

        # undo event
        self.undo = self.get_event(context, 'Screen', 'ed.undo')

        # delete event
        self.delete = self.get_event(context, 'Object Mode', 'object.delete')

        """
        # provide abstration between user and addon
        # with different select mouse side
        mouse_right = context.window_manager.keyconfigs.active.preferences.select_mouse
        if mouse_right == 'LEFT':
            mouse_left = 'RIGHT'
            mouse_right_side = 'Left'
            mouse_left_side = 'Right'
        else:
            mouse_left = 'LEFT'
            mouse_right_side = 'Right'
            mouse_left_side = 'Left'

        self.leftmouse = mouse_left + 'MOUSE'
        self.rightmouse = mouse_right + 'MOUSE'
        """

    def check(self, event, against):
        res = False
        signature = (event.alt, event.ctrl, event.shift, event.type, event.value)
        for ev in against:
            # print ("check %s == %s" % (signature, ev))
            if ev['event'] == signature:
                # print("check True")
                res = True
                break
        return res

    def get_event(self, context, keyconfig, keymap_item):
        """
            Return simple keymaps event signature as array of dict
            NOTE:
                this won't work for complex keymaps such as select_all
                using properties to call operator in different manner
            type: keyboard main type
            name: event name as defined in user preferences
            event: simple event signature to compare  like :
              if keymap.check(event, keymap.undo):
        """
        evs = [ev for k, ev in context.window_manager.keyconfigs[0].keymaps[keyconfig].keymap_items.items()
               if k == keymap_item]
        # ev = context.window_manager.keyconfigs[0].keymaps[keyconfig].keymap_items[keymap_item]
        res = []
        for ev in evs:
            key = ev.type
            if ev.ctrl:
                key += '+CTRL'
            if ev.alt:
                key += '+ALT'
            if ev.shift:
                key += '+SHIFT'
            res.append({'type': key, 'name': ev.name, 'event': (ev.alt, ev.ctrl, ev.shift, ev.type, ev.value)})
        return res

    def dump_keys(self, context, filename="/tmp/keymap.txt"):
        """
            Utility for developers :
            Dump all keymaps to a file
            filename : string a file path to dump keymaps
        """
        str = ""
        kms = context.window_manager.keyconfigs
        for name, km in kms.items():
            for key in km.keymaps.keys():
                str += "\n\n#--------------------------------\n{} - {}:\n#--------------------------------\n\n".format(name, key)
                for sub in km[key].keymap_items.keys():
                    k = km[key].keymap_items[sub]
                    str += "alt:{} ctrl:{} shift:{} type:{} value:{}  idname:{} name:{}\n".format(
                        k.alt, k.ctrl, k.shift, k.type, k.value, sub, k.name)
        file = open(filename, "w")
        file.write(str)
        file.close()



# gpl authors: nfloyd, Francesco Siddi


import logging
module_logger = logging.getLogger(__name__)

import hashlib
import bpy


#Utility function get preferences setting for exporters
def get_preferences():
    # replace the key if the add-on name changes
    addon = bpy.context.preferences.addons[__package__]
    show_warn = (addon.preferences.show_exporters if addon else False)

    return show_warn


class StoredView():
    def __init__(self, mode, index=None):
        self.logger = logging.getLogger('%s.StoredView' % __name__)
        self.scene = bpy.context.scene
        self.view3d = bpy.context.space_data
        self.index = index
        self.data_store = DataStore(mode=mode)

    def save(self):
        if self.index == -1:
            stored_view, self.index = self.data_store.create()
        else:
            stored_view = self.data_store.get(self.index)
        self.from_v3d(stored_view)
        self.logger.debug('index: %s name: %s' % (self.data_store.current_index, stored_view.name))

    def set(self):
        stored_view = self.data_store.get(self.index)
        self.update_v3d(stored_view)
        self.logger.debug('index: %s name: %s' % (self.data_store.current_index, stored_view.name))

    def from_v3d(self, stored_view):
        raise NotImplementedError("Subclass must implement abstract method")

    def update_v3d(self, stored_view):
        raise NotImplementedError("Subclass must implement abstract method")

    @staticmethod
    def is_modified(context, stored_view):
        raise NotImplementedError("Subclass must implement abstract method")


class POV(StoredView):
    def __init__(self, index=None):
        super().__init__(mode='POV', index=index)
        self.logger = logging.getLogger('%s.POV' % __name__)

    def from_v3d(self, stored_view):
        view3d = self.view3d
        region3d = view3d.region_3d

        stored_view.distance = region3d.view_distance
        stored_view.location = region3d.view_location
        stored_view.rotation = region3d.view_rotation
        stored_view.perspective_matrix_md5 = POV._get_perspective_matrix_md5(region3d)
        stored_view.perspective = region3d.view_perspective
        stored_view.lens = view3d.lens
        stored_view.clip_start = view3d.clip_start
        stored_view.clip_end = view3d.clip_end

        if region3d.view_perspective == 'CAMERA':
            stored_view.camera_type = view3d.camera.type  # type : 'CAMERA' or 'MESH'
            stored_view.camera_name = view3d.camera.name  # store string instead of object
        if view3d.lock_object is not None:
            stored_view.lock_object_name = view3d.lock_object.name  # idem
        else:
            stored_view.lock_object_name = ""
        stored_view.lock_cursor = view3d.lock_cursor
        stored_view.cursor_location = bpy.context.scene.cursor.location

    def update_v3d(self, stored_view):
        view3d = self.view3d
        region3d = view3d.region_3d
        region3d.view_distance = stored_view.distance
        region3d.view_location = stored_view.location
        region3d.view_rotation = stored_view.rotation
        region3d.view_perspective = stored_view.perspective
        view3d.lens = stored_view.lens
        view3d.clip_start = stored_view.clip_start
        view3d.clip_end = stored_view.clip_end
        view3d.lock_cursor = stored_view.lock_cursor
        if stored_view.lock_cursor is True:
            # update cursor only if view is locked to cursor
            view3d.cursor_location = stored_view.cursor_location

        if stored_view.perspective == "CAMERA":

            lock_obj = self._get_object(stored_view.lock_object_name)
            if lock_obj:
                view3d.lock_object = lock_obj
            else:
                cam = self._get_object(stored_view.camera_name)
                if cam:
                    view3d.camera = cam

    @staticmethod
    def _get_object(name, pointer=None):
        return bpy.data.objects.get(name)

    @staticmethod
    def is_modified(context, stored_view):
        # TODO: check for others param, currently only perspective
        # and perspective_matrix are checked
        POV.logger = logging.getLogger('%s.POV' % __name__)
        view3d = context.space_data
        region3d = view3d.region_3d
        if region3d.view_perspective != stored_view.perspective:
            POV.logger.debug('view_perspective')
            return True

        md5 = POV._get_perspective_matrix_md5(region3d)
        if (md5 != stored_view.perspective_matrix_md5 and
          region3d.view_perspective != "CAMERA"):
            POV.logger.debug('perspective_matrix')
            return True

        return False

    @staticmethod
    def _get_perspective_matrix_md5(region3d):
        md5 = hashlib.md5(str(region3d.perspective_matrix).encode('utf-8')).hexdigest()
        return md5


# class Layers(StoredView):
#     def __init__(self, index=None):
#         super().__init__(mode='COLLECTIONS', index=index)
#         self.logger = logging.getLogger('%s.Layers' % __name__)

#     def from_v3d(self, stored_view):
#         view3d = self.view3d
#         stored_view.view_layers = bpy.types.layer.collection
#         stored_view.scene_layers = self.scene.layers
#         stored_view.lock_camera = view3d.lock_camera

#     def update_v3d(self, stored_view):
#         view3d = self.view3d
#         view3d.camera = stored_view.camera
#         if stored_view.camera is True:
#             self.scene.layers = stored_view.scene_layers
#         else:
#             view3d.layers = stored_view.view_layers

    # @staticmethod
    # def is_modified(context, stored_view):
    #     Layers.logger = logging.getLogger('%s.Layers' % __name__)
    #     if stored_view.camera != context.space_data.camera:
    #         Layers.logger.debug('lock_camera')
    #         return True
    #     if stored_view.camera is True:
    #         for i in range(20):
    #             if stored_view.scene_layers[i] != context.scene.layers[i]:
    #                 Layers.logger.debug('scene_layers[%s]' % (i, ))
    #                 return True
    #     else:
    #         for i in range(20):
    #             if stored_view.view_layers[i] != context.space_data.view3d.layers[i]:
    #                 return True
    #     return False


# class Display(StoredView):
#     def __init__(self, index=None):
#         super().__init__(mode='DISPLAY', index=index)
#         self.logger = logging.getLogger('%s.Display' % __name__)

#     def from_v3d(self, stored_view):
#         view3d = self.view3d
#         stored_view.viewport_shade = view3d.viewport_shade
#         stored_view.show_only_render = view3d.show_only_render
#         stored_view.show_outline_selected = view3d.show_outline_selected
#         stored_view.show_all_objects_origin = view3d.show_all_objects_origin
#         stored_view.show_relationship_lines = view3d.show_relationship_lines
#         stored_view.show_floor = view3d.show_floor
#         stored_view.show_axis_x = view3d.show_axis_x
#         stored_view.show_axis_y = view3d.show_axis_y
#         stored_view.show_axis_z = view3d.show_axis_z
#         stored_view.grid_lines = view3d.grid_lines
#         stored_view.grid_scale = view3d.grid_scale
#         stored_view.grid_subdivisions = view3d.grid_subdivisions
#         stored_view.material_mode = self.scene.game_settings.material_mode
#         stored_view.show_textured_solid = view3d.show_textured_solid

    # def update_v3d(self, stored_view):
    #     view3d = self.view3d
    #     view3d.viewport_shade = stored_view.viewport_shade
    #     view3d.show_only_render = stored_view.show_only_render
    #     view3d.show_outline_selected = stored_view.show_outline_selected
    #     view3d.show_all_objects_origin = stored_view.show_all_objects_origin
    #     view3d.show_relationship_lines = stored_view.show_relationship_lines
    #     view3d.show_floor = stored_view.show_floor
    #     view3d.show_axis_x = stored_view.show_axis_x
    #     view3d.show_axis_y = stored_view.show_axis_y
    #     view3d.show_axis_z = stored_view.show_axis_z
    #     view3d.grid_lines = stored_view.grid_lines
    #     view3d.grid_scale = stored_view.grid_scale
    #     view3d.grid_subdivisions = stored_view.grid_subdivisions
    #     self.scene.game_settings.material_mode = stored_view.material_mode
    #     view3d.show_textured_solid = stored_view.show_textured_solid

    # @staticmethod
    # def is_modified(context, stored_view):
    #     Display.logger = logging.getLogger('%s.Display' % __name__)
    #     view3d = context.space_data
    #     excludes = ["material_mode", "quad_view", "lock_rotation", "show_sync_view", "use_box_clip", "name"]
    #     for k, v in stored_view.items():
    #         if k not in excludes:
    #             if getattr(view3d, k) != getattr(stored_view, k):
    #                 return True

    #     if stored_view.material_mode != context.scene.game_settings.material_mode:
    #         Display.logger.debug('material_mode')
    #         return True


class View(StoredView):
    def __init__(self, index=None):
        super().__init__(mode='VIEW', index=index)
        self.logger = logging.getLogger('%s.View' % __name__)
        self.pov = POV()
        # self.layers = Layers()
        # self.display = Display()

    def from_v3d(self, stored_view):
        self.pov.from_v3d(stored_view.pov)
        # self.layers.from_v3d(stored_view.layers)
        # self.display.from_v3d(stored_view.display)

    def update_v3d(self, stored_view):
        self.pov.update_v3d(stored_view.pov)
        # self.layers.update_v3d(stored_view.layers)
        # self.display.update_v3d(stored_view.display)

    @staticmethod
    def is_modified(context, stored_view):
        POV.is_modified(context, stored_view.pov) #or \
        #    Layers.is_modified(context, stored_view.layers) or \
        #    Display.is_modified(context, stored_view.display):
        return True
        # return False


class DataStore():
    def __init__(self, scene=None, mode=None):
        if scene is None:
            scene = bpy.context.scene
        stored_views = scene.stored_views
        self.mode = mode

        if mode is None:
            self.mode = stored_views.mode

        if self.mode == 'VIEW':
            self.list = stored_views.view_list
            self.current_index = stored_views.current_indices[0]
        elif self.mode == 'POV':
            self.list = stored_views.pov_list
            self.current_index = stored_views.current_indices[1]
        elif self.mode == 'LAYERS':
            self.list = stored_views.layers_list
            self.current_index = stored_views.current_indices[2]
        elif self.mode == 'DISPLAY':
            self.list = stored_views.display_list
            self.current_index = stored_views.current_indices[3]

    def create(self):
        item = self.list.add()
        item.name = self._generate_name()
        index = len(self.list) - 1
        self._set_current_index(index)
        return item, index

    def get(self, index):
        self._set_current_index(index)
        return self.list[index]

    def delete(self, index):
        if self.current_index > index:
            self._set_current_index(self.current_index - 1)
        elif self.current_index == index:
            self._set_current_index(-1)

        self.list.remove(index)

    def _set_current_index(self, index):
        self.current_index = index
        mode = self.mode
        stored_views = bpy.context.scene.stored_views
        if mode == 'VIEW':
            stored_views.current_indices[0] = index
        elif mode == 'POV':
            stored_views.current_indices[1] = index
        elif mode == 'LAYERS':
            stored_views.current_indices[2] = index
        elif mode == 'DISPLAY':
            stored_views.current_indices[3] = index

    def _generate_name(self):
        default_name = str(self.mode)
        names = []
        for i in self.list:
            i_name = i.name
            if i_name.startswith(default_name):
                names.append(i_name)
        names.sort()
        try:
            l_name = names[-1]
            post_fix = l_name.rpartition('.')[2]
            if post_fix.isnumeric():
                post_fix = str(int(post_fix) + 1).zfill(3)
            else:
                if post_fix == default_name:
                    post_fix = "001"
            return default_name + "." + post_fix
        except:
            return default_name

    @staticmethod
    def sanitize_data(scene):

        def check_objects_references(mode, list):
            to_remove = []
            for i, list_item in enumerate(list.items()):
                key, item = list_item
                if mode == 'POV' or mode == 'VIEWS':
                    if mode == 'VIEWS':
                        item = item.pov

                    if item.perspective == "CAMERA":

                        camera = bpy.data.objects.get(item.camera_name)
                        if camera is None:
                            try:  # pick a default camera TODO: ask to pick?
                                camera = bpy.data.cameras[0]
                                item.camera_name = camera.name
                            except:  # couldn't find a camera in the scene
                                pass

                        obj = bpy.data.objects.get(item.lock_object_name)
                        if obj is None and camera is None:
                            to_remove.append(i)

            for i in reversed(to_remove):
                list.remove(i)

        modes = ['POV', 'VIEW', 'DISPLAY', 'LAYERS']
        for mode in modes:
            data = DataStore(scene=scene, mode=mode)
            check_objects_references(mode, data.list)


def stored_view_factory(mode, *args, **kwargs):
    if mode == 'POV':
        return POV(*args, **kwargs)
    elif mode == 'LAYERS':
        return Layers(*args, **kwargs)
    elif mode == 'DISPLAY':
        return Display(*args, **kwargs)
    elif mode == 'VIEW':
        return View(*args, **kwargs)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>
import bpy
from bpy.types import Menu, Panel, UIList
from rna_prop_ui import PropertyPanel
from bpy_extras.node_utils import find_node_input


class MATERIAL_MT_context_menu(Menu):
    bl_label = "Material Specials"

    def draw(self, _context):
        layout = self.layout

        layout.operator("material.copy", icon='COPYDOWN')
        layout.operator("object.material_slot_copy")
        layout.operator("material.paste", icon='PASTEDOWN')
        layout.operator("object.material_slot_remove_unused")


class MATERIAL_UL_matslots(UIList):

    def draw_item(self, _context, layout, _data, item, icon, _active_data, _active_propname, _index):
        # assert(isinstance(item, bpy.types.MaterialSlot)
        # ob = data
        slot = item
        ma = slot.material

        layout.context_pointer_set("id", ma)

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            if ma:
                layout.prop(ma, "name", text="", emboss=False, icon_value=icon)
            else:
                layout.label(text="", icon_value=icon)
        elif self.layout_type == 'GRID':
            layout.alignment = 'CENTER'
            layout.label(text="", icon_value=icon)


class MaterialButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "material"
    # COMPAT_ENGINES must be defined in each subclass, external engines can add themselves here

    @classmethod
    def poll(cls, context):
        mat = context.material
        return mat and (context.engine in cls.COMPAT_ENGINES) and not mat.grease_pencil


class MATERIAL_PT_preview(MaterialButtonsPanel, Panel):
    bl_label = "Preview"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_EEVEE'}

    def draw(self, context):
        self.layout.template_preview(context.material)


class MATERIAL_PT_custom_props(MaterialButtonsPanel, PropertyPanel, Panel):
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}
    _context_path = "material"
    _property_type = bpy.types.Material


class EEVEE_MATERIAL_PT_context_material(MaterialButtonsPanel, Panel):
    bl_label = ""
    bl_context = "material"
    bl_options = {'HIDE_HEADER'}
    COMPAT_ENGINES = {'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    @classmethod
    def poll(cls, context):
        ob = context.object
        mat = context.material

        if (ob and ob.type == 'GPENCIL') or (mat and mat.grease_pencil):
            return False

        return (ob or mat) and (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout

        mat = context.material
        ob = context.object
        slot = context.material_slot
        space = context.space_data

        if ob:
            is_sortable = len(ob.material_slots) > 1
            rows = 3
            if is_sortable:
                rows = 5

            row = layout.row()

            row.template_list("MATERIAL_UL_matslots", "", ob, "material_slots", ob, "active_material_index", rows=rows)

            col = row.column(align=True)
            col.operator("object.material_slot_add", icon='ADD', text="")
            col.operator("object.material_slot_remove", icon='REMOVE', text="")

            col.separator()

            col.menu("MATERIAL_MT_context_menu", icon='DOWNARROW_HLT', text="")

            if is_sortable:
                col.separator()

                col.operator("object.material_slot_move", icon='TRIA_UP', text="").direction = 'UP'
                col.operator("object.material_slot_move", icon='TRIA_DOWN', text="").direction = 'DOWN'

        row = layout.row()

        if ob:
            row.template_ID(ob, "active_material", new="material.new")

            if slot:
                icon_link = 'MESH_DATA' if slot.link == 'DATA' else 'OBJECT_DATA'
                row.prop(slot, "link", icon=icon_link, icon_only=True)

            if ob.mode == 'EDIT':
                row = layout.row(align=True)
                row.operator("object.material_slot_assign", text="Assign")
                row.operator("object.material_slot_select", text="Select")
                row.operator("object.material_slot_deselect", text="Deselect")

        elif mat:
            row.template_ID(space, "pin_id")


def panel_node_draw(layout, ntree, _output_type, input_name):
    node = ntree.get_output_node('EEVEE')

    if node:
        input = find_node_input(node, input_name)
        if input:
            layout.template_node_view(ntree, node, input)
        else:
            layout.label(text="Incompatible output node")
    else:
        layout.label(text="No output node")


class EEVEE_MATERIAL_PT_surface(MaterialButtonsPanel, Panel):
    bl_label = "Surface"
    bl_context = "material"
    COMPAT_ENGINES = {'BLENDER_EEVEE'}

    def draw(self, context):
        layout = self.layout

        mat = context.material

        layout.prop(mat, "use_nodes", icon='NODETREE')
        layout.separator()

        layout.use_property_split = True

        if mat.use_nodes:
            panel_node_draw(layout, mat.node_tree, 'OUTPUT_MATERIAL', "Surface")
        else:
            layout.prop(mat, "diffuse_color", text="Base Color")
            layout.prop(mat, "metallic")
            layout.prop(mat, "specular_intensity", text="Specular")
            layout.prop(mat, "roughness")


class EEVEE_MATERIAL_PT_volume(MaterialButtonsPanel, Panel):
    bl_label = "Volume"
    bl_context = "material"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_EEVEE'}

    @classmethod
    def poll(cls, context):
        engine = context.engine
        mat = context.material
        return mat and mat.use_nodes and (engine in cls.COMPAT_ENGINES) and not mat.grease_pencil

    def draw(self, context):
        layout = self.layout

        layout.use_property_split = True

        mat = context.material

        panel_node_draw(layout, mat.node_tree, 'OUTPUT_MATERIAL', "Volume")


def draw_material_settings(self, context):
    layout = self.layout
    layout.use_property_split = True
    layout.use_property_decorate = False

    mat = context.material

    layout.prop(mat, "use_backface_culling")
    layout.prop(mat, "blend_method")
    layout.prop(mat, "shadow_method")

    row = layout.row()
    row.active = ((mat.blend_method == 'CLIP') or (mat.shadow_method == 'CLIP'))
    row.prop(mat, "alpha_threshold")

    if mat.blend_method not in {'OPAQUE', 'CLIP', 'HASHED'}:
        layout.prop(mat, "show_transparent_back")

    layout.prop(mat, "use_screen_refraction")
    layout.prop(mat, "refraction_depth")
    layout.prop(mat, "use_sss_translucency")
    layout.prop(mat, "pass_index")


class EEVEE_MATERIAL_PT_settings(MaterialButtonsPanel, Panel):
    bl_label = "Settings"
    bl_context = "material"
    COMPAT_ENGINES = {'BLENDER_EEVEE'}

    def draw(self, context):
        draw_material_settings(self, context)


class EEVEE_MATERIAL_PT_viewport_settings(MaterialButtonsPanel, Panel):
    bl_label = "Settings"
    bl_context = "material"
    bl_parent_id = "MATERIAL_PT_viewport"
    COMPAT_ENGINES = {'BLENDER_RENDER'}

    def draw(self, context):
        draw_material_settings(self, context)


class MATERIAL_PT_viewport(MaterialButtonsPanel, Panel):
    bl_label = "Viewport Display"
    bl_context = "material"
    bl_options = {'DEFAULT_CLOSED'}
    bl_order = 10

    @classmethod
    def poll(cls, context):
        mat = context.material
        return mat and not mat.grease_pencil

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        mat = context.material

        col = layout.column()
        col.prop(mat, "diffuse_color", text="Color")
        col.prop(mat, "metallic")
        col.prop(mat, "roughness")


class MATERIAL_PT_lineart(MaterialButtonsPanel, Panel):
    bl_label = "Line Art"
    bl_options = {'DEFAULT_CLOSED'}
    bl_order = 10

    @classmethod
    def poll(cls, context):
        mat = context.material
        return mat and not mat.grease_pencil

    def draw(self, context):
        layout = self.layout

        mat = context.material
        lineart = mat.lineart

        layout.prop(lineart, "use_transparency")

        if lineart.use_transparency:

            layout.label(text="Transparency Masks:")

            row = layout.row(align=True)
            for i in range(8):
                row.prop(lineart, "use_transparency_mask", text=str(i), index=i, toggle=True)


classes = (
    MATERIAL_MT_context_menu,
    MATERIAL_UL_matslots,
    MATERIAL_PT_preview,
    EEVEE_MATERIAL_PT_context_material,
    EEVEE_MATERIAL_PT_surface,
    EEVEE_MATERIAL_PT_volume,
    EEVEE_MATERIAL_PT_settings,
    MATERIAL_PT_lineart,
    MATERIAL_PT_viewport,
    EEVEE_MATERIAL_PT_viewport_settings,
    MATERIAL_PT_custom_props,
)


if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

"""update new variables to values from older API.

It does not have a UI and used to be found with F3 search field.
This file needs an update."""

import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
    EnumProperty,
)


def update2_0_0_9():
    # Temporally register old props, so we can access their values.
    register()

    # Mapping old names -> old default values
    # XXX We could also store the new name, but as it is just the same without leading pov_ ...
    # Get default values of pov scene props.
    old_sce_props = {}
    for k in [
        "pov_tempfiles_enable",
        "pov_deletefiles_enable",
        "pov_scene_name",
        "pov_scene_path",
        "pov_renderimage_path",
        "pov_list_lf_enable",
        "pov_radio_enable",
        "pov_radio_display_advanced",
        "pov_media_enable",
        "pov_media_samples",
        "pov_media_color",
        "pov_baking_enable",
        "pov_indentation_character",
        "pov_indentation_spaces",
        "pov_comments_enable",
        "pov_command_line_switches",
        "pov_antialias_enable",
        "pov_antialias_method",
        "pov_antialias_depth",
        "pov_antialias_threshold",
        "pov_jitter_enable",
        "pov_jitter_amount",
        "pov_antialias_gamma",
        "pov_max_trace_level",
        "pov_photon_spacing",
        "pov_photon_max_trace_level",
        "pov_photon_adc_bailout",
        "pov_photon_gather_min",
        "pov_photon_gather_max",
        "pov_radio_adc_bailout",
        "pov_radio_always_sample",
        "pov_radio_brightness",
        "pov_radio_count",
        "pov_radio_error_bound",
        "pov_radio_gray_threshold",
        "pov_radio_low_error_factor",
        "pov_radio_media",
        "pov_radio_minimum_reuse",
        "pov_radio_nearest_count",
        "pov_radio_normal",
        "pov_radio_recursion_limit",
        "pov_radio_pretrace_start",
        "pov_radio_pretrace_end",
    ]:
        old_sce_props[k] = getattr(bpy.types.Scene, k)[1].get('default', None)

    # Get default values of pov material props.
    old_mat_props = {}
    for k in [
        "pov_irid_enable",
        "pov_mirror_use_IOR",
        "pov_mirror_metallic",
        "pov_conserve_energy",
        "pov_irid_amount",
        "pov_irid_thickness",
        "pov_irid_turbulence",
        "pov_interior_fade_color",
        "pov_caustics_enable",
        "pov_fake_caustics",
        "pov_fake_caustics_power",
        "pov_photons_refraction",
        "pov_photons_dispersion",
        "pov_photons_reflection",
        "pov_refraction_type",
        "pov_replacement_text",
    ]:
        old_mat_props[k] = getattr(bpy.types.Material, k)[1].get(
            'default', None
        )

    # Get default values of pov texture props.
    old_tex_props = {}
    for k in [
        "pov_tex_gamma_enable",
        "pov_tex_gamma_value",
        "pov_replacement_text",
    ]:
        old_tex_props[k] = getattr(bpy.types.Texture, k)[1].get('default', None)

    # Get default values of pov object props.
    old_obj_props = {}
    for k in [
        "pov_importance_value",
        "pov_collect_photons",
        "pov_replacement_text",
    ]:
        old_obj_props[k] = getattr(bpy.types.Object, k)[1].get('default', None)

    # Get default values of pov camera props.
    old_cam_props = {}
    for k in [
        "pov_dof_enable",
        "pov_dof_aperture",
        "pov_dof_samples_min",
        "pov_dof_samples_max",
        "pov_dof_variance",
        "pov_dof_confidence",
        "pov_replacement_text",
    ]:
        old_cam_props[k] = getattr(bpy.types.Camera, k)[1].get('default', None)

    # Get default values of pov text props.
    old_txt_props = {}
    for k in ["pov_custom_code"]:
        old_txt_props[k] = getattr(bpy.types.Text, k)[1].get('default', None)

    ################################################################################################
    # Now, update !
    # For each old pov property of each scene, if its value is not equal to the default one,
    # copy it to relevant new prop...
    for sce in bpy.data.scenes:
        for k, d in old_sce_props.items():
            val = getattr(sce, k, d)
            if val != d:
                setattr(sce.pov, k[4:], val)
    # The same goes for materials, textures, etc.
    for mat in bpy.data.materials:
        for k, d in old_mat_props.items():
            val = getattr(mat, k, d)
            if val != d:
                setattr(mat.pov, k[4:], val)
    for tex in bpy.data.textures:
        for k, d in old_tex_props.items():
            val = getattr(tex, k, d)
            if val != d:
                setattr(tex.pov, k[4:], val)
    for obj in bpy.data.objects:
        for k, d in old_obj_props.items():
            val = getattr(obj, k, d)
            if val != d:
                setattr(obj.pov, k[4:], val)
    for cam in bpy.data.cameras:
        for k, d in old_cam_props.items():
            val = getattr(cam, k, d)
            if val != d:
                setattr(cam.pov, k[4:], val)
    for txt in bpy.data.texts:
        for k, d in old_txt_props.items():
            val = getattr(txt, k, d)
            if val != d:
                setattr(txt.pov, k[4:], val)
    # Finally, unregister old props !
    unregister()


class RenderCopySettings(bpy.types.Operator):
    """Update old POV properties to new ones"""

    bl_idname = "scene.pov_update_properties"
    bl_label = "PovRay render: Update to script v0.0.9"
    bl_option = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        update2_0_0_9()
        return {'FINISHED'}


def register():
    Scene = bpy.types.Scene
    Mat = bpy.types.Material
    Tex = bpy.types.Texture
    Obj = bpy.types.Object
    Cam = bpy.types.Camera
    Text = bpy.types.Text
    ###########################SCENE##################################

    # File Options
    Scene.pov_tempfiles_enable = BoolProperty(
        name="Enable Tempfiles",
        description="Enable the OS-Tempfiles. Otherwise set the path where to save the files",
        default=True,
    )
    Scene.pov_deletefiles_enable = BoolProperty(
        name="Delete files",
        description="Delete files after rendering. Doesn't work with the image",
        default=True,
    )
    Scene.pov_scene_name = StringProperty(
        name="Scene Name",
        description="Name of POV-Ray scene to create. Empty name will use the name of the blend file",
        default="",
        maxlen=1024,
    )
    Scene.pov_scene_path = StringProperty(
        name="Export scene path",
        # description="Path to directory where the exported scene (POV and INI) is created",  # Bug in POV-Ray RC3
        description="Path to directory where the files are created",
        default="",
        maxlen=1024,
        subtype="DIR_PATH",
    )
    Scene.pov_renderimage_path = StringProperty(
        name="Rendered image path",
        description="Full path to directory where the rendered image is saved",
        default="",
        maxlen=1024,
        subtype="DIR_PATH",
    )
    Scene.pov_list_lf_enable = BoolProperty(
        name="LF in lists",
        description="Enable line breaks in lists (vectors and indices). Disabled: lists are exported in one line",
        default=True,
    )

    # Not a real pov option, just to know if we should write
    Scene.pov_radio_enable = BoolProperty(
        name="Enable Radiosity",
        description="Enable POV-Rays radiosity calculation",
        default=False,
    )
    Scene.pov_radio_display_advanced = BoolProperty(
        name="Advanced Options",
        description="Show advanced options",
        default=False,
    )
    Scene.pov_media_enable = BoolProperty(
        name="Enable Media",
        description="Enable POV-Rays atmospheric media",
        default=False,
    )
    Scene.pov_media_samples = IntProperty(
        name="Samples",
        description="Number of samples taken from camera to first object encountered along ray path for media calculation",
        min=1,
        max=100,
        default=35,
    )

    Scene.pov_media_color = FloatVectorProperty(
        name="Media Color",
        description="The atmospheric media color",
        subtype='COLOR',
        precision=4,
        step=0.01,
        min=0,
        soft_max=1,
        default=(0.001, 0.001, 0.001),
        options={'ANIMATABLE'},
    )

    Scene.pov_baking_enable = BoolProperty(
        name="Enable Baking",
        description="Enable POV-Rays texture baking",
        default=False,
    )
    Scene.pov_indentation_character = EnumProperty(
        name="Indentation",
        description="Select the indentation type",
        items=(
            ("0", "None", "No indentation"),
            ("1", "Tabs", "Indentation with tabs"),
            ("2", "Spaces", "Indentation with spaces"),
        ),
        default="2",
    )
    Scene.pov_indentation_spaces = IntProperty(
        name="Quantity of spaces",
        description="The number of spaces for indentation",
        min=1,
        max=10,
        default=4,
    )

    Scene.pov_comments_enable = BoolProperty(
        name="Enable Comments",
        description="Add comments to pov file",
        default=True,
    )

    # Real pov options
    Scene.pov_command_line_switches = StringProperty(
        name="Command Line Switches",
        description="Command line switches consist of a + (plus) or - (minus) sign, followed by one or more alphabetic characters and possibly a numeric value",
        default="",
        maxlen=500,
    )

    Scene.pov_antialias_enable = BoolProperty(
        name="Anti-Alias", description="Enable Anti-Aliasing", default=True
    )

    Scene.pov_antialias_method = EnumProperty(
        name="Method",
        description="AA-sampling method. Type 1 is an adaptive, non-recursive, super-sampling method. Type 2 is an adaptive and recursive super-sampling method",
        items=(
            ("0", "non-recursive AA", "Type 1 Sampling in POV-Ray"),
            ("1", "recursive AA", "Type 2 Sampling in POV-Ray"),
        ),
        default="1",
    )

    Scene.pov_antialias_depth = IntProperty(
        name="Antialias Depth",
        description="Depth of pixel for sampling",
        min=1,
        max=9,
        default=3,
    )

    Scene.pov_antialias_threshold = FloatProperty(
        name="Antialias Threshold",
        description="Tolerance for sub-pixels",
        min=0.0,
        max=1.0,
        soft_min=0.05,
        soft_max=0.5,
        default=0.1,
    )

    Scene.pov_jitter_enable = BoolProperty(
        name="Jitter",
        description="Enable Jittering. Adds noise into the sampling process (it should be avoided to use jitter in animation)",
        default=True,
    )

    Scene.pov_jitter_amount = FloatProperty(
        name="Jitter Amount",
        description="Amount of jittering",
        min=0.0,
        max=1.0,
        soft_min=0.01,
        soft_max=1.0,
        default=1.0,
    )

    Scene.pov_antialias_gamma = FloatProperty(
        name="Antialias Gamma",
        description="POV-Ray compares gamma-adjusted values for super sampling. Antialias Gamma sets the Gamma before comparison",
        min=0.0,
        max=5.0,
        soft_min=0.01,
        soft_max=2.5,
        default=2.5,
    )

    Scene.pov_max_trace_level = IntProperty(
        name="Max Trace Level",
        description="Number of reflections/refractions allowed on ray path",
        min=1,
        max=256,
        default=5,
    )

    Scene.pov_photon_spacing = FloatProperty(
        name="Spacing",
        description="Average distance between photons on surfaces. half this get four times as many surface photons",
        min=0.001,
        max=1.000,
        soft_min=0.001,
        soft_max=1.000,
        default=0.005,
        precision=3,
    )

    Scene.pov_photon_max_trace_level = IntProperty(
        name="Max Trace Level",
        description="Number of reflections/refractions allowed on ray path",
        min=1,
        max=256,
        default=5,
    )

    Scene.pov_photon_adc_bailout = FloatProperty(
        name="ADC Bailout",
        description="The adc_bailout for photons. Use adc_bailout = 0.01 / brightest_ambient_object for good results",
        min=0.0,
        max=1000.0,
        soft_min=0.0,
        soft_max=1.0,
        default=0.1,
        precision=3,
    )

    Scene.pov_photon_gather_min = IntProperty(
        name="Gather Min",
        description="Minimum number of photons gathered for each point",
        min=1,
        max=256,
        default=20,
    )

    Scene.pov_photon_gather_max = IntProperty(
        name="Gather Max",
        description="Maximum number of photons gathered for each point",
        min=1,
        max=256,
        default=100,
    )

    Scene.pov_radio_adc_bailout = FloatProperty(
        name="ADC Bailout",
        description="The adc_bailout for radiosity rays. Use adc_bailout = 0.01 / brightest_ambient_object for good results",
        min=0.0,
        max=1000.0,
        soft_min=0.0,
        soft_max=1.0,
        default=0.01,
        precision=3,
    )

    Scene.pov_radio_always_sample = BoolProperty(
        name="Always Sample",
        description="Only use the data from the pretrace step and not gather any new samples during the final radiosity pass",
        default=True,
    )

    Scene.pov_radio_brightness = FloatProperty(
        name="Brightness",
        description="Amount objects are brightened before being returned upwards to the rest of the system",
        min=0.0,
        max=1000.0,
        soft_min=0.0,
        soft_max=10.0,
        default=1.0,
    )

    Scene.pov_radio_count = IntProperty(
        name="Ray Count",
        description="Number of rays for each new radiosity value to be calculated (halton sequence over 1600)",
        min=1,
        max=10000,
        soft_max=1600,
        default=35,
    )

    Scene.pov_radio_error_bound = FloatProperty(
        name="Error Bound",
        description="One of the two main speed/quality tuning values, lower values are more accurate",
        min=0.0,
        max=1000.0,
        soft_min=0.1,
        soft_max=10.0,
        default=1.8,
    )

    Scene.pov_radio_gray_threshold = FloatProperty(
        name="Gray Threshold",
        description="One of the two main speed/quality tuning values, lower values are more accurate",
        min=0.0,
        max=1.0,
        soft_min=0,
        soft_max=1,
        default=0.0,
    )

    Scene.pov_radio_low_error_factor = FloatProperty(
        name="Low Error Factor",
        description="Just enough samples is slightly blotchy. Low error changes error tolerance for less critical last refining pass",
        min=0.0,
        max=1.0,
        soft_min=0.0,
        soft_max=1.0,
        default=0.5,
    )

    # max_sample - not available yet
    Scene.pov_radio_media = BoolProperty(
        name="Media",
        description="Radiosity estimation can be affected by media",
        default=False,
    )

    Scene.pov_radio_minimum_reuse = FloatProperty(
        name="Minimum Reuse",
        description="Fraction of the screen width which sets the minimum radius of reuse for each sample point (At values higher than 2% expect errors)",
        min=0.0,
        max=1.0,
        soft_min=0.1,
        soft_max=0.1,
        default=0.015,
        precision=3,
    )

    Scene.pov_radio_nearest_count = IntProperty(
        name="Nearest Count",
        description="Number of old ambient values blended together to create a new interpolated value",
        min=1,
        max=20,
        default=5,
    )

    Scene.pov_radio_normal = BoolProperty(
        name="Normals",
        description="Radiosity estimation can be affected by normals",
        default=False,
    )

    Scene.pov_radio_recursion_limit = IntProperty(
        name="Recursion Limit",
        description="how many recursion levels are used to calculate the diffuse inter-reflection",
        min=1,
        max=20,
        default=3,
    )

    Scene.pov_radio_pretrace_start = FloatProperty(
        name="Pretrace Start",
        description="Fraction of the screen width which sets the size of the blocks in the mosaic preview first pass",
        min=0.01,
        max=1.00,
        soft_min=0.02,
        soft_max=1.0,
        default=0.08,
    )

    Scene.pov_radio_pretrace_end = FloatProperty(
        name="Pretrace End",
        description="Fraction of the screen width which sets the size of the blocks in the mosaic preview last pass",
        min=0.001,
        max=1.00,
        soft_min=0.01,
        soft_max=1.00,
        default=0.04,
        precision=3,
    )

    #############################MATERIAL######################################

    Mat.pov_irid_enable = BoolProperty(
        name="Enable Iridescence",
        description="Newton's thin film interference (like an oil slick on a puddle of water or the rainbow hues of a soap bubble.)",
        default=False,
    )

    Mat.pov_mirror_use_IOR = BoolProperty(
        name="Correct Reflection",
        description="Use same IOR as raytrace transparency to calculate mirror reflections. More physically correct",
        default=False,
    )

    Mat.pov_mirror_metallic = BoolProperty(
        name="Metallic Reflection",
        description="mirror reflections get colored as diffuse (for metallic materials)",
        default=False,
    )

    Mat.pov_conserve_energy = BoolProperty(
        name="Conserve Energy",
        description="Light transmitted is more correctly reduced by mirror reflections, also the sum of diffuse and translucency gets reduced below one ",
        default=True,
    )

    Mat.pov_irid_amount = FloatProperty(
        name="amount",
        description="Contribution of the iridescence effect to the overall surface color. As a rule of thumb keep to around 0.25 (25% contribution) or less, but experiment. If the surface is coming out too white, try lowering the diffuse and possibly the ambient values of the surface",
        min=0.0,
        max=1.0,
        soft_min=0.01,
        soft_max=1.0,
        default=0.25,
    )

    Mat.pov_irid_thickness = FloatProperty(
        name="thickness",
        description="A very thin film will have a high frequency of color changes while a thick film will have large areas of color",
        min=0.0,
        max=1000.0,
        soft_min=0.1,
        soft_max=10.0,
        default=1,
    )

    Mat.pov_irid_turbulence = FloatProperty(
        name="turbulence",
        description="This parameter varies the thickness",
        min=0.0,
        max=10.0,
        soft_min=0.000,
        soft_max=1.0,
        default=0,
    )

    Mat.pov_interior_fade_color = FloatVectorProperty(
        name="Fade Color",
        description="Color of filtered attenuation for transparent materials",
        subtype='COLOR',
        precision=4,
        step=0.01,
        min=0.0,
        soft_max=1.0,
        default=(0, 0, 0),
        options={'ANIMATABLE'},
    )

    Mat.pov_caustics_enable = BoolProperty(
        name="Caustics",
        description="use only fake refractive caustics (default) or photon based reflective/refractive caustics",
        default=True,
    )

    Mat.pov_fake_caustics = BoolProperty(
        name="Fake Caustics",
        description="use only (Fast) fake refractive caustics",
        default=True,
    )

    Mat.pov_fake_caustics_power = FloatProperty(
        name="Fake caustics power",
        description="Values typically range from 0.0 to 1.0 or higher. Zero is no caustics. Low, non-zero values give broad hot-spots while higher values give tighter, smaller simulated focal points",
        min=0.00,
        max=10.0,
        soft_min=0.00,
        soft_max=1.10,
        default=0.1,
    )

    Mat.pov_photons_refraction = BoolProperty(
        name="Refractive Photon Caustics",
        description="more physically correct",
        default=False,
    )

    Mat.pov_photons_dispersion = FloatProperty(
        name="chromatic dispersion",
        description="Light passing through will be separated according to wavelength. This ratio of refractive indices for violet to red controls how much the colors are spread out 1 = no dispersion, good values are 1.01 to 1.1",
        min=1.0000,
        max=10.000,
        soft_min=1.0000,
        soft_max=1.1000,
        precision=4,
        default=1.0000,
    )

    Mat.pov_photons_reflection = BoolProperty(
        name="Reflective Photon Caustics",
        description="Use this to make your Sauron's ring ;-P",
        default=False,
    )

    Mat.pov_refraction_type = EnumProperty(
        items=[
            ("0", "None", "use only reflective caustics"),
            ("1", "Fake Caustics", "use fake caustics"),
            ("2", "Photons Caustics", "use photons for refractive caustics"),
        ],
        name="Refractive",
        description="use fake caustics (fast) or true photons for refractive Caustics",
        default="1",
    )
    ##################################CustomPOV Code############################
    Mat.pov_replacement_text = StringProperty(
        name="Declared name:",
        description="Type the declared name in custom POV code or an external .inc it points at. texture {} expected",
        default="",
    )

    # Only DUMMIES below for now:
    Tex.pov_replacement_text = StringProperty(
        name="Declared name:",
        description="Type the declared name in custom POV code or an external .inc it points at. pigment {} expected",
        default="",
    )

    Obj.pov_replacement_text = StringProperty(
        name="Declared name:",
        description="Type the declared name in custom POV code or an external .inc it points at. Any POV shape expected e.g: isosurface {}",
        default="",
    )

    Cam.pov_replacement_text = StringProperty(
        name="Texts in blend file",
        description="Type the declared name in custom POV code or an external .inc it points at. camera {} expected",
        default="",
    )
    ##############################TEXTURE######################################

    # Custom texture gamma
    Tex.pov_tex_gamma_enable = BoolProperty(
        name="Enable custom texture gamma",
        description="Notify some custom gamma for which texture has been precorrected without the file format carrying it and only if it differs from your OS expected standard (see pov doc)",
        default=False,
    )

    Tex.pov_tex_gamma_value = FloatProperty(
        name="Custom texture gamma",
        description="value for which the file was issued e.g. a Raw photo is gamma 1.0",
        min=0.45,
        max=5.00,
        soft_min=1.00,
        soft_max=2.50,
        default=1.00,
    )

    #################################OBJECT####################################

    # Importance sampling
    Obj.pov_importance_value = FloatProperty(
        name="Radiosity Importance",
        description="Priority value relative to other objects for sampling radiosity rays. Increase to get more radiosity rays at comparatively small yet bright objects",
        min=0.01,
        max=1.00,
        default=1.00,
    )

    # Collect photons
    Obj.pov_collect_photons = BoolProperty(
        name="Receive Photon Caustics",
        description="Enable object to collect photons from other objects caustics. Turn off for objects that don't really need to receive caustics (e.g. objects that generate caustics often don't need to show any on themselves) ",
        default=True,
    )

    ##################################CAMERA###################################

    # DOF Toggle
    Cam.pov_dof_enable = BoolProperty(
        name="Depth Of Field",
        description="Enable POV-Ray Depth Of Field ",
        default=True,
    )

    # Aperture (Intensity of the Blur)
    Cam.pov_dof_aperture = FloatProperty(
        name="Aperture",
        description="Similar to a real camera's aperture effect over focal blur (though not in physical units and independent of focal length).Increase to get more blur",
        min=0.01,
        max=1.00,
        default=0.25,
    )

    # Aperture adaptive sampling
    Cam.pov_dof_samples_min = IntProperty(
        name="Samples Min",
        description="Minimum number of rays to use for each pixel",
        min=1,
        max=128,
        default=96,
    )

    Cam.pov_dof_samples_max = IntProperty(
        name="Samples Max",
        description="Maximum number of rays to use for each pixel",
        min=1,
        max=128,
        default=128,
    )

    Cam.pov_dof_variance = IntProperty(
        name="Variance",
        description="Minimum threshold (fractional value) for adaptive DOF sampling (up increases quality and render time). The value for the variance should be in the range of the smallest displayable color difference",
        min=1,
        max=100000,
        soft_max=10000,
        default=256,
    )

    Cam.pov_dof_confidence = FloatProperty(
        name="Confidence",
        description="Probability to reach the real color value. Larger confidence values will lead to more samples, slower traces and better images",
        min=0.01,
        max=0.99,
        default=0.90,
    )

    ###################################TEXT####################################

    Text.pov_custom_code = BoolProperty(
        name="Custom Code",
        description="Add this text at the top of the exported POV-Ray file",
        default=False,
    )


def unregister():
    Scene = bpy.types.Scene
    Mat = bpy.types.Material
    Tex = bpy.types.Texture
    Obj = bpy.types.Object
    Cam = bpy.types.Camera
    Text = bpy.types.Text
    del Scene.pov_tempfiles_enable  # CR
    del Scene.pov_scene_name  # CR
    del Scene.pov_deletefiles_enable  # CR
    del Scene.pov_scene_path  # CR
    del Scene.pov_renderimage_path  # CR
    del Scene.pov_list_lf_enable  # CR
    del Scene.pov_radio_enable
    del Scene.pov_radio_display_advanced
    del Scene.pov_radio_adc_bailout
    del Scene.pov_radio_always_sample
    del Scene.pov_radio_brightness
    del Scene.pov_radio_count
    del Scene.pov_radio_error_bound
    del Scene.pov_radio_gray_threshold
    del Scene.pov_radio_low_error_factor
    del Scene.pov_radio_media
    del Scene.pov_radio_minimum_reuse
    del Scene.pov_radio_nearest_count
    del Scene.pov_radio_normal
    del Scene.pov_radio_recursion_limit
    del Scene.pov_radio_pretrace_start  # MR
    del Scene.pov_radio_pretrace_end  # MR
    del Scene.pov_media_enable  # MR
    del Scene.pov_media_samples  # MR
    del Scene.pov_media_color  # MR
    del Scene.pov_baking_enable  # MR
    del Scene.pov_max_trace_level  # MR
    del Scene.pov_photon_spacing  # MR
    del Scene.pov_photon_max_trace_level  # MR
    del Scene.pov_photon_adc_bailout  # MR
    del Scene.pov_photon_gather_min  # MR
    del Scene.pov_photon_gather_max  # MR
    del Scene.pov_antialias_enable  # CR
    del Scene.pov_antialias_method  # CR
    del Scene.pov_antialias_depth  # CR
    del Scene.pov_antialias_threshold  # CR
    del Scene.pov_antialias_gamma  # CR
    del Scene.pov_jitter_enable  # CR
    del Scene.pov_jitter_amount  # CR
    del Scene.pov_command_line_switches  # CR
    del Scene.pov_indentation_character  # CR
    del Scene.pov_indentation_spaces  # CR
    del Scene.pov_comments_enable  # CR
    del Mat.pov_irid_enable  # MR
    del Mat.pov_mirror_use_IOR  # MR
    del Mat.pov_mirror_metallic  # MR
    del Mat.pov_conserve_energy  # MR
    del Mat.pov_irid_amount  # MR
    del Mat.pov_irid_thickness  # MR
    del Mat.pov_irid_turbulence  # MR
    del Mat.pov_interior_fade_color  # MR
    del Mat.pov_caustics_enable  # MR
    del Mat.pov_fake_caustics  # MR
    del Mat.pov_fake_caustics_power  # MR
    del Mat.pov_photons_refraction  # MR
    del Mat.pov_photons_dispersion  # MR
    del Mat.pov_photons_reflection  # MR
    del Mat.pov_refraction_type  # MR
    del Mat.pov_replacement_text  # MR
    del Tex.pov_tex_gamma_enable  # MR
    del Tex.pov_tex_gamma_value  # MR
    del Tex.pov_replacement_text  # MR
    del Obj.pov_importance_value  # MR
    del Obj.pov_collect_photons  # MR
    del Obj.pov_replacement_text  # MR
    del Cam.pov_dof_enable  # MR
    del Cam.pov_dof_aperture  # MR
    del Cam.pov_dof_samples_min  # MR
    del Cam.pov_dof_samples_max  # MR
    del Cam.pov_dof_variance  # MR
    del Cam.pov_dof_confidence  # MR
    del Cam.pov_replacement_text  # MR
    del Text.pov_custom_code  # MR



"""
Note on Altering Data
+++++++++++++++++++++

Altering data from handlers should be done carefully. While rendering the
``frame_change_pre`` and ``frame_change_post`` handlers are called from one
thread and the viewport updates from a different thread. If the handler changes
data that is accessed by the viewport, this can cause a crash of Blender. In
such cases, lock the interface (Render → Lock Interface or
:data:`bpy.types.RenderSettings.use_lock_interface`) before starting a render.

Below is an example of a mesh that is altered from a handler:
"""

def frame_change_pre(scene):
    # A triangle that shifts in the z direction
    zshift = scene.frame_current * 0.1
    vertices = [(-1, -1, zshift), (1, -1, zshift), (0, 1, zshift)]
    triangles = [(0, 1, 2)]

    object = bpy.data.objects["The Object"]
    object.data.clear_geometry()
    object.data.from_pydata(vertices, [], triangles)



#
# Copyright (C) 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# This file is part of Power Sequencer.
#
# Power Sequencer is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Power Sequencer is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Power Sequencer. If
# not, see <https://www.gnu.org/licenses/>.
#
import bpy


class InfoProgressBar:
    """
    Draws a progress bar in the info header area
    """

    def __init__(self, progress_min=0, progress_max=100):
        assert progress_min < progress_max

        self.progress_min = progress_min
        self.progress_max = progress_max

        self._progress = bpy.props.FloatProperty(
            default=self.progress_min,
            min=self.progress_min,
            max=self.progress_max,
            subtype="PERCENTAGE",
        )
        self._visible = False

    def update(self, context):
        for area in context.screen.areas:
            if area.type == "INFO":
                area.tag_redraw()

    def draw(self):
        if self.progress >= self.progress_max:
            self.visible = False
        else:
            self.layout.prop(self, "_progress", text="Progress", slider=True)

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = min(max(self.progress_min, value), self.progress_max)

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, value):
        self._visible = value

        if self._visible:
            bpy.types.INFO_HT_header.append(self.draw)
            bpy.app.handlers.scene_update_post.add(self.update)
        else:
            bpy.types.INFO_HT_header.remove(self.draw)
            bpy.app.handlers.scene_update_post.remove(self.update)
